"""A utility to specify axis ranges."""
import sys
from math import inf, isfinite
from typing import Callable, Final

import numpy as np
from matplotlib.axes import Axes  # type: ignore
from pycommons.types import type_error

from moptipy.api.logging import (
    KEY_BEST_F,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    TIME_UNIT_FES,
    TIME_UNIT_MILLIS,
)
from moptipy.evaluation.end_statistics import KEY_ERT_FES, KEY_ERT_TIME_MILLIS

#: The internal minimum float value for log-scaled axes.
_MIN_LOG_FLOAT: Final[float] = sys.float_info.min


class AxisRanger:
    """An object for simplifying axis range computations."""

    def __init__(self,
                 chosen_min: float | None = None,
                 chosen_max: float | None = None,
                 use_data_min: bool = True,
                 use_data_max: bool = True,
                 log_scale: bool = False,
                 log_base: float | None = None):
        """
        Initialize the axis ranger.

        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: use the minimum found in the data?
        :param use_data_max:  use the maximum found in the data?
        :param log_scale: should the axis be log-scaled?
        :param log_base: the base to be used for the logarithm
        """
        if not isinstance(log_scale, bool):
            raise type_error(log_scale, "log_scale", bool)
        #: Should the axis be log-scaled?
        self.log_scale: Final[bool] = log_scale

        self.__log_base: Final[float | None] = \
            log_base if self.log_scale else None
        if self.__log_base is not None:
            if not isinstance(log_base, float):
                raise type_error(log_base, "log_base", float)
            if log_base <= 1.0:
                raise ValueError(f"log_base must be > 1, but is {log_base}.")

        if chosen_min is not None:
            if not isinstance(chosen_min, float | int):
                raise type_error(chosen_min, "chosen_min", (int, float))
            chosen_min = float(chosen_min)
            if not isfinite(chosen_min):
                raise ValueError(f"chosen_min cannot be {chosen_min}.")
            if self.log_scale and (chosen_min <= 0):
                raise ValueError(
                    f"if log_scale={self.log_scale}, then chosen_min must "
                    f"be > 0, but is {chosen_min}.")

        #: The pre-defined, chosen minimum axis value.
        self.__chosen_min: Final[float | None] = chosen_min

        if chosen_max is not None:
            if not isinstance(chosen_max, float | int):
                raise type_error(chosen_max, "chosen_max", (int, float))
            chosen_max = float(chosen_max)
            if not isfinite(chosen_max):
                raise ValueError(f"chosen_max cannot be {chosen_max}.")
            if (self.__chosen_min is not None) and \
                    (chosen_max <= self.__chosen_min):
                raise ValueError(f"If chosen_min is {self.__chosen_min}, then"
                                 f" chosen_max cannot be {chosen_max}.")

        #: The pre-defined, chosen maximum axis value.
        self.__chosen_max: Final[float | None] = chosen_max

        if not isinstance(use_data_min, bool):
            raise type_error(use_data_min, "use_data_min", bool)
        #: Should we use the data min value?
        self.__use_data_min: Final[bool] = use_data_min

        if not isinstance(use_data_max, bool):
            raise type_error(use_data_max, "use_data_max", bool)
        #: Should we use the data max value?
        self.__use_data_max: Final[bool] = use_data_max

        #: The minimum detected from the data.
        self.__detected_min: float = inf

        #: The maximum detected from the data.
        self.__detected_max: float = _MIN_LOG_FLOAT if self.log_scale \
            else -inf

        #: Did we detect a minimum?
        self.__has_detected_min = False
        #: Did we detect a maximum?
        self.__has_detected_max = False

    def register_array(self, data: np.ndarray) -> None:
        """
        Register a data array.

        :param data: the data to register
        """
        if self.__use_data_min or self.__use_data_max:
            d = data[np.isfinite(data)]
            if self.__use_data_min:
                self.register_value(float(d.min()))
            if self.__use_data_max:
                self.register_value(float(d.max()))

    def register_value(self, value: float) -> None:
        """
        Register a single value.

        :param value: the data to register
        """
        if isfinite(value):
            if self.__use_data_min and (
                    (value < self.__detected_min)
                    and ((value > 0.0) or (not self.log_scale))):
                self.__detected_min = value
                self.__has_detected_min = True
            if self.__use_data_max and (value > self.__detected_max):
                self.__detected_max = value
                self.__has_detected_max = True

    def pad_detected_range(self, pad_min: bool = False,
                           pad_max: bool = False) -> None:
        """
        Add some padding to the current detected range.

        This function increases the current detected or chosen maximum value
        and/or decreases the current detected minimum by a small amount. This
        can be useful when we want to plot stuff that otherwise would become
        invisible because it would be directly located at the boundary of a
        plot.

        This function works by computing a slightly smaller/larger value than
        the current detected minimum/maximum and then passing it to
        :meth:`register_value`. It can only work if the end(s) chosen for
        padding are in "detect" mode and the other end is either in "detect"
        or "chosen" mode.

        This method should be called *only* once and *only* after all data has
        been registered (via :meth:`register_value` :meth:`register_array`)
        and before calling :meth:`apply`.

        :param pad_min: should we pad the minimum?
        :param pad_max: should we pad the maximum?

        :raises ValueError: if this axis ranger is not configured to use a
            detected minimum/maximum or does not have a detected
            minimum/maximum or any other invalid situation occurs
        """
        if not isinstance(pad_min, bool):
            raise type_error(pad_min, "pad_min", bool)
        if not isinstance(pad_max, bool):
            raise type_error(pad_max, "pad_max", bool)
        if not (pad_min or pad_max):
            return

        max_value: float
        min_value: float
        if self.__use_data_min:
            if not self.__has_detected_min:
                raise ValueError("No minimum detected so far.")
            min_value = self.__detected_min
        else:
            if pad_min:
                raise ValueError("Can only pad minimum if use_data_min.")
            if self.__chosen_min is None:
                raise ValueError("Chosen min is None!")
            min_value = self.__chosen_min

        if self.__use_data_max:
            if not self.__has_detected_max:
                raise ValueError("No maximum detected so far.")
            max_value = self.__detected_max
        else:
            if pad_max:
                raise ValueError("Can only pad maximum if use_data_max.")
            if self.__chosen_max is None:
                raise ValueError("Chosen max is None!")
            max_value = self.__chosen_max

        if min_value >= max_value:
            raise ValueError(
                f"minimum={min_value} while maximum={max_value}.")

        new_max: float
        if pad_max:
            if max_value >= inf:
                return
            new_max = max_value + (3.0 * (max_value - min_value)) / 100.0
            if not isfinite(new_max) or (new_max <= max_value):
                raise ValueError(f"invalid padded max={new_max} at min="
                                 f"{min_value} and max={max_value}.")
            self.register_value(new_max)
        else:
            new_max = max_value

        new_min: float
        if pad_min:
            if min_value <= -inf:
                return
            new_min = min_value - (3.0 * (max_value - min_value)) / 100.0
            if self.log_scale and (new_min <= 0.0 < min_value):
                new_min = 0.5 * min_value
            if not isfinite(new_min) or (new_min >= min_value):
                raise ValueError(f"invalid padded min={new_min} at min="
                                 f"{min_value} and max={max_value}.")
            self.register_value(new_min)
        else:
            new_min = min_value

        if new_min > new_max:
            raise ValueError(f"new_min={new_min}, new_max={new_max}??")

    def apply(self, axes: Axes, which_axis: str) -> None:
        """
        Apply this axis ranger to the given axis.

        :param axes: the axes object to which the ranger shall be applied
        :param which_axis: the axis to which it should be applied, either
            `"x"` or `"y"` or both (`"xy"`)
        """
        if not isinstance(which_axis, str):
            raise type_error(which_axis, "which_axis", str)

        for is_x_axis in (True, False):
            if ("x" if is_x_axis else "y") not in which_axis:
                continue

            use_min, use_max = \
                axes.get_xlim() if is_x_axis else axes.get_ylim()

            if not isfinite(use_min):
                raise ValueError(f"Minimum data interval cannot be {use_min}.")
            if not isfinite(use_max):
                raise ValueError(f"Maximum data interval cannot be {use_max}.")
            if use_max <= use_min:
                raise ValueError(f"Invalid axis range[{use_min},{use_max}].")

            replace_range = False

            if self.__chosen_min is not None:
                use_min = self.__chosen_min
                replace_range = True
            elif self.__use_data_min:
                if not self.__has_detected_min:
                    raise ValueError("No minimum in data detected.")
                use_min = self.__detected_min
                replace_range = True

            if self.__chosen_max is not None:
                use_max = self.__chosen_max
                replace_range = True
            elif self.__use_data_max:
                if not self.__has_detected_max:
                    raise ValueError("No maximum in data detected.")
                use_max = self.__detected_max
                replace_range = True

            if replace_range:
                if use_min >= use_max:
                    raise ValueError(
                        f"Invalid computed range [{use_min},{use_max}].")
                if is_x_axis:
                    axes.set_xlim(use_min, use_max)
                else:
                    axes.set_ylim(use_min, use_max)

            if self.log_scale:
                if use_min <= 0:
                    raise ValueError("minimum must be positive if log scale "
                                     f"is defined, but found {use_min}.")
                if is_x_axis:
                    if self.__log_base is None:
                        axes.semilogx()
                    else:
                        axes.semilogx(base=self.__log_base)
                elif self.__log_base is None:
                    axes.semilogy()
                else:
                    axes.semilogy(base=self.__log_base)

    def get_pinf_replacement(self) -> float:
        """
        Get a reasonable finite value that can replace positive infinity.

        :return: a reasonable finite value that can be used to replace
            positive infinity
        """
        data_max: float = 0.0
        if self.__chosen_max is not None:
            data_max = self.__chosen_max
        elif self.__has_detected_max:
            data_max = self.__detected_max
        return min(1e100, max(1e70, 1e5 * data_max))

    def get_0_replacement(self) -> float:
        """
        Get a reasonable positive finite value that can replace `0`.

        :return: a reasonable finite value that can be used to replace
            `0`
        """
        data_min: float = 1e-100
        if self.__chosen_min is not None:
            data_min = self.__chosen_min
        elif self.__has_detected_min:
            data_min = self.__detected_min
        return max(1e-100, min(1e-70, 1e-5 * data_min))

    @staticmethod
    def for_axis(name: str,
                 chosen_min: float | None = None,
                 chosen_max: float | None = None,
                 use_data_min: bool | None = None,
                 use_data_max: bool | None = None,
                 log_scale: bool | None = None,
                 log_base: float | None = None) -> "AxisRanger":
        """
        Create a default axis ranger based on the axis type.

        The axis ranger will use the minimal values and log scaling options
        that usually make sense for the dimension, unless overridden by the
        optional arguments.

        :param name: the axis type name, supporting `"ms"`, `"FEs"`,
            `"plainF"`, `"scaledF"`, and `"normalizedF"`
        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: should the data minimum be used
        :param use_data_max: should the data maximum be used
        :param log_scale: the log scale indicator
        :param log_base: the log base
        :return: the `AxisRanger`
        """
        if not isinstance(name, str):
            raise type_error(name, "axis name", str)

        l_log: bool = False
        l_min: float | None = None
        l_max: float | None = None
        l_data_min: bool = chosen_min is None
        l_data_max: bool = chosen_max is None

        if name in {TIME_UNIT_MILLIS, KEY_LAST_IMPROVEMENT_TIME_MILLIS,
                    KEY_TOTAL_TIME_MILLIS, KEY_ERT_TIME_MILLIS}:
            if chosen_min is not None:
                if (chosen_min < 0) or (not isfinite(chosen_min)):
                    raise ValueError("chosen_min must be >= 0 for axis "
                                     f"type {name}, but is {chosen_min}.")
                l_log = (chosen_min > 0)
                if log_scale is not None:
                    if log_scale and (not l_log):
                        raise ValueError(f"Cannot set log_scale={log_scale} "
                                         f"and chosen_min={chosen_min} for "
                                         f"axis type {name}.")
                    l_log = log_scale
            elif log_scale is None:
                l_log = True
            else:
                l_log = log_scale

            l_min = (1 if l_log else 0) if chosen_min is None else chosen_min

            if use_data_max is not None:
                l_data_max = use_data_max

            l_data_min = False if use_data_min is None else use_data_min

            if chosen_max is not None:
                l_max = chosen_max

            return AxisRanger(l_min, l_max, l_data_min, l_data_max,
                              l_log, log_base if l_log else None)

        if name in {TIME_UNIT_FES, KEY_LAST_IMPROVEMENT_FE, KEY_TOTAL_FES,
                    KEY_ERT_FES}:
            if chosen_min is None:
                l_min = 1
            else:
                if (chosen_min < 1) or (not isfinite(chosen_min)):
                    raise ValueError("chosen_min must be >= 1 for axis "
                                     f"type {name}, but is {chosen_min}.")
                l_min = chosen_min
            l_log = True if (log_scale is None) else log_scale

            if use_data_max is not None:
                l_data_max = use_data_max

            l_data_min = False if use_data_min is None else use_data_min

            return AxisRanger(l_min, chosen_max, l_data_min, l_data_max,
                              l_log, log_base if l_log else None)

        if name in {F_NAME_RAW, KEY_BEST_F}:
            if use_data_max is not None:
                l_data_max = use_data_max
            if use_data_min is not None:
                l_data_min = use_data_min
            if log_scale is not None:
                l_log = log_scale
            return AxisRanger(chosen_min, chosen_max, l_data_min, l_data_max,
                              l_log, log_base if l_log else None)

        if name == F_NAME_SCALED:
            l_min = 1
        elif name == F_NAME_NORMALIZED:
            if (log_scale is None) or (not log_scale):
                l_min = 0
        elif name == "ecdf":
            if (log_scale is None) or (not log_scale):
                l_min = 0
            l_max = 1
        else:
            raise ValueError(f"Axis type {name!r} is unknown.")

        if chosen_min is not None:
            l_min = chosen_min

        if log_scale is not None:
            l_log = log_scale
        if use_data_max is not None:
            l_data_max = use_data_max
        if use_data_min is not None:
            l_data_min = use_data_min

        return AxisRanger(l_min, chosen_max, l_data_min, l_data_max,
                          l_log, log_base if l_log else None)

    @staticmethod
    def for_axis_func(chosen_min: float | None = None,
                      chosen_max: float | None = None,
                      use_data_min: bool | None = None,
                      use_data_max: bool | None = None,
                      log_scale: bool | None = None,
                      log_base: float | None = None) -> Callable:
        """
        Generate a function that provides the default per-axis ranger.

        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: should the data minimum be used
        :param use_data_max: should the data maximum be used
        :param log_scale: the log scale indicator
        :param log_base: the log base
        :return: a function in the shape of :meth:`for_axis` with the
            provided defaults
        """
        def __func(name: str,
                   cmi=chosen_min,
                   cma=chosen_max,
                   udmi=use_data_min,
                   udma=use_data_max,
                   ls=log_scale,
                   lb=log_base) -> AxisRanger:
            return AxisRanger.for_axis(name, cmi, cma, udmi, udma, ls, lb)

        return __func



"""
A tool for selecting a consistent subset of data from partial experiments.

When we have partial experimental data, maybe collected from experiments that
are still ongoing, we want to still evaluate them in some consistent way. The
right method for doing this could be to select a subset of that data that is
consistent, i.e., a subset where the algorithms have the same number of runs
on the instances using the same seeds. The function :func:`select_consistent`
offered by this module provides the functionality to make such a selection.
It may be a bit slow, but hopefully it will pick the largest possible
consistent sub-selection or, at least, get close to it.

The current method to select the data is rather heuristic. It always begins
with the full set of data and aims to delete the element that will cause the
least other deletions down the road, until we arrive in a consistent state.
I strongly suspect that doing this perfectly would be NP-hard, so we cannot
implement this. Instead, we use different heuristics and then pick the best
result.
"""

from collections import Counter
from operator import itemgetter
from typing import Any, Callable, Final, Iterable, TypeVar

from pycommons.io.console import logger
from pycommons.types import type_error

from moptipy.evaluation.base import PerRunData

#: the type variable for the selector routine
T = TypeVar("T", bound=PerRunData)

#: the algorithm key
KEY_ALGORITHM: Final[int] = 0
#: the instance key
KEY_INSTANCE: Final[int] = KEY_ALGORITHM + 1
#: the encoding key
KEY_ENCODING: Final[int] = KEY_INSTANCE + 1
#: the objective key
KEY_OBJECTIVE: Final[int] = KEY_ENCODING + 1
#: the seed key
KEY_SEED: Final[int] = KEY_OBJECTIVE + 1
#: the number of keys
TOTAL_KEYS: Final[int] = KEY_SEED + 1


def __data_tup(d: PerRunData) -> tuple[str, str, str, str, tuple[str, int]]:
    """
    Get a raw data tuple for the given record.

    :param d: the data record
    :returns: the data tuple
    """
    if not isinstance(d, PerRunData):
        raise type_error(d, "dataElement", PerRunData)
    return d.algorithm, d.instance, d.encoding, d.objective, (
        d.instance, d.rand_seed)


def __score_inc(scores: tuple[Counter, ...], key: tuple[Any, ...]) -> None:
    """
    Increment the score of a given tuple.

    :param scores: the scores
    :param key: the key
    """
    for i, k in enumerate(key):
        scores[i][k] += 1


def __score_dec(scores: tuple[Counter, ...], key: tuple[Any, ...]) -> None:
    """
    Decrement the score of a given tuple.

    :param scores: the scores
    :param key: the key
    """
    for i, k in enumerate(key):
        scores[i][k] -= 1
        z: int = scores[i][k]
        if z <= 0:
            del scores[i][k]
            if z < 0:
                raise ValueError("Got negative score?")


def __ret(source: list, expected: int, log: bool) -> list:
    """
    Prepare a source list for return.

    :param source: the list
    :param expected: the number of expected records
    :param log: shall we log information
    :returns: the result
    """
    for i, er in enumerate(source):
        source[i] = er[1]
    if list.__len__(source) != expected:
        raise ValueError("Inconsistent list length!")
    if log:
        logger(f"Now returning {expected} records of data.")
    return source  # type: ignore


def __scorer_tired(d: tuple, scores: tuple[Counter, ...], _) -> list[int]:
    """
    Score based on the tired score.

    :param d: the tuple to score
    :param scores: the scores
    :returns: the score
    """
    return sorted(scores[i][t] for i, t in enumerate(d))


def __scorer_normalized_tired(d: tuple, scores: tuple[Counter, ...],
                              total: tuple[int, ...]) -> list[float]:
    """
    Score based on the tired score.

    :param d: the tuple to score
    :param scores: the scores
    :param total: the total of the scores
    :returns: the score
    """
    return sorted(
        scores[i][t] / total[i] for i, t in enumerate(d) if total[i] > 0)


def __scorer_sum(d: tuple, scores: tuple[Counter, ...], _) -> int:
    """
    Score based on the sum score.

    :param d: the tuple to score
    :param scores: the scores
    :returns: the score
    """
    return sum(scores[i][t] for i, t in enumerate(d))


def __scorer_normalized_sum(d: tuple, scores: tuple[Counter, ...],
                            total: tuple[int, ...]) -> float:
    """
    Score based on the sum score.

    :param d: the tuple to score
    :param scores: the scores
    :param total: the total of the scores
    :returns: the score
    """
    return sum(
        scores[i][t] / total[i] for i, t in enumerate(d) if total[i] > 0)


def __scorer_product(d: tuple, scores: tuple[Counter, ...], _) -> int:
    """
    Score based on the product score.

    :param d: the tuple to score
    :param scores: the scores
    :returns: the score
    """
    result: int = 1
    for i, t in enumerate(d):
        result *= scores[i][t]
    return result


def __scorer_normalized_min(d: tuple, scores: tuple[Counter, ...],
                            total: tuple[int, ...]) -> float:
    """
    Score based on the normalized minimum score.

    :param d: the tuple to score
    :param scores: the scores
    :param total: the total of the scores
    :returns: the score
    """
    return min(
        scores[i][t] / total[i] for i, t in enumerate(d) if total[i] > 0)


def __not(_: tuple[Counter, ...]) -> int:
    """
    Do nothing, return a placeholder.

    :returns -1: always
    """
    return -1


def __total(scores: tuple[Counter, ...]) -> tuple[int, ...] | None:
    """
    Get the total scores.

    :param scores: the scores
    :returns: the total scores
    """
    result: tuple[int, ...] = tuple(
        t.total() if len(t) > 1 else -1 for t in scores)
    return None if (tuple.__len__(result) < 1) or (
        max(result) <= 0) else result


#: the scorers
__SCORERS: Final[tuple[tuple[str, Callable[[
    tuple, tuple[Counter, ...], Any], Any], Callable[[
        tuple[Counter, ...]], Any]], ...]] = (
    ("tired", __scorer_tired, __not),
    ("normalized_tired", __scorer_normalized_tired, __total),
    ("sum", __scorer_sum, __not),
    ("normalized_sum", __scorer_normalized_sum, __total),
    ("product", __scorer_product, __total),
    ("normalized_min", __scorer_normalized_min, __total),
)


def select_consistent(data: Iterable[T], log: bool = True,
                      thorough: bool = True) -> list[T]:
    """
    Select data such that the numbers of runs are consistent.

    The input is a set of data items which represent some records over the
    runs of algorithms on instances. It may be that not all algorithms have
    been applied to all instances. Maybe the number of runs is inconsistent
    over the algorithm-instance combinations, too. Maybe some algorithms
    have more runs on some instances. Maybe the runs are even different,
    it could be that some algorithms have runs for seed `A`, `B`, and `C` on
    instance `I`, while others have runs for seed `C` and `D`. This function
    is designed to retain only the runs with seed `C` in such a case. It may
    discard algorithms or instances or algorithm-instance-seed combinations
    in order to obtain a selection of data where all algorithms have been
    applied to all instances as same as often and using the same seeds.

    Now there are different ways to select such consistent subsets of a
    dataset. Of course, we want to select the data such that as much as
    possible of the data is retained and as little as possible is discarded.
    This may be a hard optimization problem in itself. Here, we offer a
    heuristic solution. Basically, we step-by-step try to cut away the
    setups that are covered by the least amount of runs. We keep repeating
    this until we arrive in a situation where all setups have the same
    amount of runs. We then check if there were some strange symmetries that
    still make the data inconsistent and, if we found some, try to delete
    one run to break the symmetries and then repeat the cleaning-up process.
    In the end, we should get a list of overall consistent data elements that
    can be used during a normal experiment evaluation procedure.

    This iterative process may be rather slow on larger datasets, but it is
    maybe the best approximation we can offer to retain as much data as
    possible.

    :param data: the source data
    :param log: shall we log the progress
    :param thorough: use the slower method which may give us more data
    :return: a list with the selected data

    >>> def __p(x) -> str:
    ...     return (f"{x.algorithm}/{x.instance}/{x.objective}/{x.encoding}/"
    ...             f"{x.rand_seed}")

    >>> a1i1o1e1s1 = PerRunData("a1", "i1", "o1", "e1", 1)
    >>> a1i1o1e1s2 = PerRunData("a1", "i1", "o1", "e1", 2)
    >>> a1i1o1e1s3 = PerRunData("a1", "i1", "o1", "e1", 3)
    >>> a1i2o1e1s1 = PerRunData("a1", "i2", "o1", "e1", 1)
    >>> a1i2o1e1s2 = PerRunData("a1", "i2", "o1", "e1", 2)
    >>> a1i2o1e1s3 = PerRunData("a1", "i2", "o1", "e1", 3)
    >>> a2i1o1e1s1 = PerRunData("a2", "i1", "o1", "e1", 1)
    >>> a2i1o1e1s2 = PerRunData("a2", "i1", "o1", "e1", 2)
    >>> a2i1o1e1s3 = PerRunData("a2", "i1", "o1", "e1", 3)
    >>> a2i2o1e1s1 = PerRunData("a1", "i2", "o1", "e1", 1)
    >>> a2i2o1e1s2 = PerRunData("a2", "i2", "o1", "e1", 2)
    >>> a2i2o1e1s3 = PerRunData("a2", "i2", "o1", "e1", 3)

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a1i1o1e1s2, a1i1o1e1s3,
    ...     a1i2o1e1s1, a1i2o1e1s2, a1i2o1e1s3,
    ...     a2i1o1e1s1, a2i1o1e1s2,
    ...     a2i2o1e1s2, a2i2o1e1s3))))
    ['a1/i1/o1/e1/1', 'a1/i1/o1/e1/2', 'a1/i2/o1/e1/2', 'a1/i2/o1/e1/3',\
 'a2/i1/o1/e1/1', 'a2/i1/o1/e1/2', 'a2/i2/o1/e1/2', 'a2/i2/o1/e1/3']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s2, a1i1o1e1s3,
    ...     a1i2o1e1s1, a1i2o1e1s2, a1i2o1e1s3,
    ...     a2i1o1e1s1, a2i1o1e1s2,
    ...     a2i2o1e1s2, a2i2o1e1s3))))
    ['a1/i2/o1/e1/2', 'a1/i2/o1/e1/3', 'a2/i2/o1/e1/2', 'a2/i2/o1/e1/3']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s2, a1i1o1e1s3,
    ...     a1i2o1e1s1, a1i2o1e1s2, a1i2o1e1s3,
    ...     a2i1o1e1s1, a2i1o1e1s2,
    ...     a2i2o1e1s2))))
    ['a1/i1/o1/e1/2', 'a1/i2/o1/e1/2', 'a2/i1/o1/e1/2', 'a2/i2/o1/e1/2']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a1i1o1e1s2, a1i1o1e1s3,
    ...     a2i2o1e1s1, a2i2o1e1s2, a2i2o1e1s3))))
    ['a1/i1/o1/e1/1', 'a1/i1/o1/e1/2', 'a1/i1/o1/e1/3']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a1i1o1e1s2, a1i1o1e1s3,
    ...     a2i1o1e1s1, a2i1o1e1s2, a2i1o1e1s3))))
    ['a1/i1/o1/e1/1', 'a1/i1/o1/e1/2', 'a1/i1/o1/e1/3', \
'a2/i1/o1/e1/1', 'a2/i1/o1/e1/2', 'a2/i1/o1/e1/3']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a1i1o1e1s2, a1i2o1e1s2, a1i2o1e1s3))))
    ['a1/i1/o1/e1/1', 'a1/i1/o1/e1/2', 'a1/i2/o1/e1/2', 'a1/i2/o1/e1/3']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a1i1o1e1s2, a1i2o1e1s2))))
    ['a1/i1/o1/e1/1', 'a1/i1/o1/e1/2']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a2i1o1e1s2))))
    ['a1/i1/o1/e1/1']

    >>> list(map(__p, select_consistent((
    ...     a1i1o1e1s1, a2i1o1e1s2, a2i1o1e1s3))))
    ['a2/i1/o1/e1/2', 'a2/i1/o1/e1/3']

    >>> try:
    ...     select_consistent((a1i1o1e1s1, a1i1o1e1s2, a1i2o1e1s2, a1i2o1e1s2))
    ... except ValueError as ve:
    ...     print(ve)
    Found 4 records but only 3 different keys!

    >>> try:
    ...     select_consistent(1)
    ... except TypeError as te:
    ...     print(te)
    data should be an instance of typing.Iterable but is int, namely 1.

    >>> try:
    ...     select_consistent((a2i1o1e1s2, a2i1o1e1s3), 3)
    ... except TypeError as te:
    ...     print(te)
    log should be an instance of bool but is int, namely 3.

    >>> try:
    ...     select_consistent({234})
    ... except TypeError as te:
    ...     print(te)
    dataElement should be an instance of moptipy.evaluation.base.PerRunData \
but is int, namely 234.

    >>> try:
    ...     select_consistent((a2i1o1e1s2, a2i1o1e1s3), True, 4)
    ... except TypeError as te:
    ...     print(te)
    thorough should be an instance of bool but is int, namely 4.
    """
    if not isinstance(data, Iterable):
        raise type_error(data, "data", Iterable)
    if not isinstance(log, bool):
        raise type_error(log, "log", bool)
    if not isinstance(thorough, bool):
        raise type_error(thorough, "thorough", bool)

    # We obtain a sorted list of the data in order to make the results
    # consistent regardless of the order in which the data comes in.
    # The data is only sorted by its features and not by any other information
    # attached to it.
    dataset: Final[list[list]] = [[__data_tup(t), t, 0] for t in data]
    dataset.sort(key=itemgetter(0))
    dataset_size: int = list.__len__(dataset)
    if log:
        logger(f"Found {dataset_size} records of data.")

    set_l: Final[int] = set.__len__({x[0] for x in dataset})
    if set_l != dataset_size:
        raise ValueError(
            f"Found {dataset_size} records but only {set_l} different keys!")

    # Compute the scores: count how often each algorithm, instance,
    # objective, encoding, and (instance, seed) combination are
    # encountered.
    key_tuple_len: Final[int] = tuple.__len__(dataset[0][0])
    if key_tuple_len != TOTAL_KEYS:
        raise ValueError("Unexpected error!")

    # the map for the score calculations
    scores: Final[tuple[Counter, ...]] = tuple(
        Counter() for _ in range(key_tuple_len))

    # the final result
    best_length: int = -1
    best_data: list[list] | None = None

    for scorer_name, scorer, scorer_total in __SCORERS if thorough else (
            __SCORERS[0], ):
        logger(f"Now scoring according to the {scorer_name!r} method.")
        # compute the original item scores
        for sc in scores:
            sc.clear()

        source: list[list] = dataset.copy()
        count: int = dataset_size
        for er in source:
            __score_inc(scores, er[0])

        changed: bool = True
        while changed:
            changed = False

            # Find the setups with maximum and minimum scores.
            min_score: Any = None
            max_score: Any = None
            norm: Any = scorer_total(scores)

            if norm is None:  # cannot proceed
                if log:
                    logger(f"Method {scorer_name!r} is not applicable.")
                count = -1  # we can do nothing
                break

            for er in source:
                score = scorer(er[0], scores, norm)
                er[2] = score
                if (min_score is None) or (min_score > score):
                    min_score = score
                if (max_score is None) or (max_score < score):
                    max_score = score
            del norm

            if min_score < max_score:  # Some setups have lower scores.
                del_count: int = 0  # We will delete them.
                for i in range(count - 1, -1, -1):
                    er = source[i]
                    if er[2] <= min_score:
                        del source[i]
                        __score_dec(scores, er[0])
                        del_count += 1
                if del_count <= 0:
                    raise ValueError(
                        f"Did not delete anything under {scorer_name!r}?")

                new_count: int = list.__len__(source)
                if new_count != (count - del_count):
                    raise ValueError("List inconsistent after deletion "
                                     f"under {scorer_name!r}?")
                if log:
                    logger(
                        f"Deleted {del_count} of the {count} records because "
                        f"their score was {min_score} while the maximum score"
                        f" was {max_score}. Retained {new_count} records "
                        f"under {scorer_name!r}.")
                count = new_count
                changed = True
                continue

            if log:
                logger(f"All setups now have the same score {max_score} under"
                       f" {scorer_name!r}.")
            if count <= best_length:
                if log:
                    logger(f"We now only have {count} setups, which means we "
                           "cannot get better than the current best set with "
                           f"{best_length} setups, so we quit after score-"
                           f"based cleaning under {scorer_name!r}.")
                count = -1
                break

            # If we get here, all elements have the same score.
            # This means that we are basically done.
            #
            # However, this may also happen if a very odd division exists in
            # the data. Maybe we have one algorithm that was applied to one
            # instance ten times and another algorithm applied to another
            # instance ten times. This data would still be inconsistent, as it
            # does not allow for any comparison.

            # Compute the different values for everything except the seeds.
            keys: tuple[int, ...] = tuple(
                v for v, s in enumerate(scores)
                if (v != KEY_SEED) and (len(s) > 1))

            if tuple.__len__(keys) > 1:
                if log:
                    logger("Now checking for inconsistencies in "
                           "algorithm/instance/objective/encoding under"
                           f" {scorer_name!r}.")
                # Only if there are different values in at least one tuple
                # dimension, we need to check for strange situations.

                # For each of (instance, algorithm, encoding, objective), we
                # must have the same number of other records.
                # If not, then we have some strange symmetric situation that
                # needs to be solved.
                per_value: dict[Any, set[Any]] = {}
                last: set[Any] | None = None
                for key in keys:
                    other_keys: tuple[int, ...] = tuple(
                        kk for kk in keys if kk != key)
                    for er in source:
                        kv = er[0][key]
                        other = tuple(er[0][ok] for ok in other_keys)
                        if kv in per_value:
                            per_value[kv].add(other)
                        else:
                            per_value[kv] = last = {other}
                    for v in per_value.values():
                        if v != last:
                            changed = True
                            break
                    per_value.clear()

                    # We just need to delete one random element. This will
                    # break the symmetry
                    if changed:
                        if log:
                            logger(
                                f"Deleting one of the {count} elements to "
                                "break the erroneous symmetry under "
                                f"{scorer_name}.")
                        er = source[-1]
                        del source[-1]
                        __score_dec(scores, er[0])
                        count -= 1
                        break
                del per_value
                del keys
                if changed:
                    continue

                if log:
                    logger("No inconsistencies in algorithm/instance/objecti"
                           f"ve/encoding found under {scorer_name!r}.")
            elif log:
                logger("No inconsistencies in algorithm/instance/objective/"
                       f"encoding possible under {scorer_name!r}.")
            if count <= best_length:
                if log:
                    logger(f"We now only have {count} setups, which means we "
                           "cannot get better than the current best set with "
                           f"{best_length} setups, so we quit after algorithm"
                           "/instance/objective/encoding cleaning under "
                           f"{scorer_name!r}.")
                count = -1
                break

            # If we get here, the only problem left could be if algorithms
            # have different seeds for the same instances. We thus need to
            # check that for each instance, all the seeds are the same.
            # Notice that such inconsistencies can only occur if different
            # seeds occurred exactly as same as often.
            if log:
                logger("Now checking for inconsistencies in instance "
                       f"seeds under {scorer_name!r}.")
            seeds: dict[str, dict[tuple[str, str, str], set[int]]] = {}
            for er in source:
                inst: str = er[1].instance
                cur: dict[tuple[str, str, str], set[int]]
                if inst in seeds:
                    cur = seeds[inst]
                else:
                    seeds[inst] = cur = {}
                kx: tuple[str, str, str] = (
                    er[1].algorithm, er[1].objective, er[1].encoding)
                if kx in cur:
                    cur[kx].add(er[1].rand_seed)
                else:
                    cur[kx] = {er[1].rand_seed}

            max_seed_insts: set[str] = set()
            max_seeds: int = -1
            min_seed_inst: str | None = None
            min_seeds: int = -1
            must_delete_from_insts: set[str] = set()
            for instance, inst_setups in seeds.items():
                kvx: set[int] | None = None
                for setup_seeds in inst_setups.values():
                    if kvx is None:
                        kvx = setup_seeds
                        seed_cnt: int = set.__len__(setup_seeds)
                        if seed_cnt >= max_seeds:
                            if seed_cnt > max_seeds:
                                max_seed_insts.clear()
                                max_seeds = seed_cnt
                            max_seed_insts.add(instance)
                        if (seed_cnt < min_seeds) or (min_seed_inst is None):
                            min_seeds = seed_cnt
                            min_seed_inst = instance
                    elif kvx != setup_seeds:
                        # We got a symmetric inconsistency
                        must_delete_from_insts.add(instance)
                        break
            if min_seeds < max_seeds:
                must_delete_from_insts.update(max_seed_insts)
            del max_seed_insts

            del_count = set.__len__(must_delete_from_insts)
            if del_count > 0:
                if log:
                    logger("Must delete one record from all of "
                           f"{must_delete_from_insts!r}.")
                for i in range(count - 1, -1, -1):
                    er = source[i]
                    instance = er[1].instance
                    if instance in must_delete_from_insts:
                        must_delete_from_insts.remove(instance)
                        del source[i]
                        __score_dec(scores, er[0])
                        changed = True
                        if set.__len__(must_delete_from_insts) <= 0:
                            break
                new_count = list.__len__(source)
                if new_count != (count - del_count):
                    raise ValueError("Error when deleting instances "
                                     f"under {scorer_name!r}.")
                count = new_count
                if not changed:
                    raise ValueError(
                        f"Seeds inconsistent under {scorer_name!r}.")
            del must_delete_from_insts

            del seeds
            if changed:
                if count <= best_length:
                    if log:
                        logger(f"We now only have {count} setups, which "
                               "means we cannot get better than the current "
                               f"best set with {best_length} setups, so we "
                               "quit after seed-based cleaning under "
                               f"{scorer_name!r}.")
                    count = -1
                    break
            elif log:
                logger(f"No seed inconsistencies under {scorer_name!r}.")
            # There should not be any problems left, but we need to check
            # again if something has changed.

        if count <= 0:
            continue  # We can do nothing here

        if count > best_length:
            logger(f"Method {scorer_name!r} yields {count} records, "
                   "which is the new best.")
            best_length = count
            best_data = source
            if best_length >= dataset_size:
                logger(f"Included all data under {scorer_name!r}, "
                       "so we can stop.")
                break
        elif log:
            logger(f"Method {scorer_name!r} yields {count} records, "
                   "which is not better than the current "
                   f"best {best_length}.")

    # We are finally finished. The scores are no longer needed.
    del scores
    if best_length <= 0:
        raise ValueError("No data found at all?")
    return __ret(best_data, best_length, log)



"""The bit-string based uniform crossover operator used for intspaces."""

from typing import Final

from moptipy.operators.bitstrings.op2_uniform import Op2Uniform as __Op2Uni

#: The uniform crossover operator for integer spaces is the same as for
#: bitstrings.
Op2Uniform: Final[type[__Op2Uni]] = __Op2Uni



"""
The TwoMax problem.

TwoMax has two optima. The local optimum at the string with all `False` bits.
The objective value of this local optimum is 1. The global optimum is the
string with all `True` bits and its objective value is 0. The worst objective
value is reached if half of the bits are `True` and the other half is `False`.
Then, the objective value will be `(n // 2) + 1`.

The TwoMax problem is based on OneMax but introduces deceptiveness in the
objective function by having a local and a global optimum. Since their basins
of attraction have the same size, a (1 + 1) EA can solve the problem in
`Omega(n ln n)` steps with probability 0.5 while otherwise needing exponential
runtime in expectation, leading to a total expected runtime in
`Omega(n ** n)`.

1. Tobias Friedrich, Francesco Quinzan, and Markus Wagner. Escaping Large
   Deceptive Basins of Attraction with Heavy-Tailed Mutation Operators.
   GECCO 2018. ACM.
   doi: https://doi.org/10.1145/3205455.3205515.
2. Clarissa Van Hoyweghen, David E. Goldberg, and Bart Naudts. From TwoMax to
   the Ising Model: Easy and Hard Symmetrical Problems. GECCO 2002.
   pp 626-633. Morgan Kaufmann.
3. Tobias Friedrich, Pietro S. Oliveto, Dirk Sudholt, and Carsten Witt.
   Analysis of Diversity-Preserving Mechanisms for Global Exploration.
   Evolutionary Computation. 17(4):455-476. 2009.
   doi: https://doi.org/10.1162/evco.2009.17.4.17401.
4. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
5. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698

This is code is part of the research work of Mr. Jiazheng ZENG (曾嘉政),
a Master's student at the Institute of Applied Optimization
(应用优化研究所) of the School of Artificial
Intelligence and Big Data (人工智能与大数据学院) at
Hefei University (合肥大学) in
Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""
from typing import Callable, Final, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def twomax(x: np.ndarray) -> int:
    """
    Compute the objective value of the TwoMax problem.

    :param x: the np array
    :return: the sum of the indices of the first two ones

    >>> twomax(np.array([True, True, True, True, True]))
    0
    >>> twomax(np.array([False, False, False, False, False]))
    1
    >>> twomax(np.array([False, True, True, False, False]))
    3
    >>> twomax(np.array([True, True, True, True, True]))
    0
    >>> twomax(np.array([False, False, False, False, False]))
    1
    >>> twomax(np.array([False, True, False, False, False]))
    2
    >>> twomax(np.array([False, False, False, False, False, False]))
    1
    >>> twomax(np.array([False, False, True, False, False, False]))
    2
    >>> twomax(np.array([False, True, False, True, False, False]))
    3
    >>> twomax(np.array([False, True, False, True, False, True]))
    4
    >>> twomax(np.array([True, False, True, False, True, True]))
    3
    >>> twomax(np.array([False, True, True, True, True, True]))
    2
    >>> twomax(np.array([True, True, True, True, True, True]))
    0

    # n = 1 and 0 true bits
    >>> twomax(np.array([0]))
    1

    # n = 1 and 1 true bit
    >>> twomax(np.array([1]))
    0

    # n = 2 and 0 true bits
    >>> twomax(np.array([0, 0]))
    1

    # n = 2 and 1 true bit
    >>> twomax(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> twomax(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> twomax(np.array([0, 1]))
    2

    # n = 2 and 2 true bits
    >>> twomax(np.array([1, 1]))
    0

    # n = 3 and 0 true bits
    >>> twomax(np.array([0, 0, 0]))
    1

    # n = 3 and 1 true bit
    >>> twomax(np.array([1, 0, 0]))
    2

    # n = 3 and 1 true bit
    >>> twomax(np.array([0, 1, 0]))
    2

    # n = 3 and 1 true bit
    >>> twomax(np.array([0, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> twomax(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> twomax(np.array([0, 1, 1]))
    2

    # n = 3 and 2 true bits
    >>> twomax(np.array([1, 1, 0]))
    2

    # n = 3 and 3 true bits
    >>> twomax(np.array([1, 1, 1]))
    0

    # n = 4 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0]))
    1

    # n = 4 and 1 true bit
    >>> twomax(np.array([1, 0, 0, 0]))
    2

    # n = 4 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 1]))
    2

    # n = 4 and 1 true bit
    >>> twomax(np.array([1, 0, 0, 0]))
    2

    # n = 4 and 2 true bits
    >>> twomax(np.array([0, 1, 1, 0]))
    3

    # n = 4 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 1]))
    3

    # n = 4 and 2 true bits
    >>> twomax(np.array([0, 0, 1, 1]))
    3

    # n = 4 and 3 true bits
    >>> twomax(np.array([1, 0, 1, 1]))
    2

    # n = 4 and 3 true bits
    >>> twomax(np.array([1, 1, 1, 0]))
    2

    # n = 4 and 3 true bits
    >>> twomax(np.array([1, 1, 0, 1]))
    2

    # n = 4 and 4 true bits
    >>> twomax(np.array([1, 1, 1, 1]))
    0

    # n = 5 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0]))
    1

    # n = 5 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 1, 0]))
    2

    # n = 5 and 1 true bit
    >>> twomax(np.array([0, 1, 0, 0, 0]))
    2

    # n = 5 and 1 true bit
    >>> twomax(np.array([1, 0, 0, 0, 0]))
    2

    # n = 5 and 2 true bits
    >>> twomax(np.array([0, 1, 1, 0, 0]))
    3

    # n = 5 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 0, 1]))
    3

    # n = 5 and 2 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0]))
    3

    # n = 5 and 3 true bits
    >>> twomax(np.array([1, 1, 0, 1, 0]))
    3

    # n = 5 and 3 true bits
    >>> twomax(np.array([0, 1, 0, 1, 1]))
    3

    # n = 5 and 3 true bits
    >>> twomax(np.array([1, 0, 1, 1, 0]))
    3

    # n = 5 and 4 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1]))
    2

    # n = 5 and 4 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1]))
    2

    # n = 5 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1]))
    2

    # n = 5 and 5 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1]))
    0

    # n = 6 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 0]))
    1

    # n = 6 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 1]))
    2

    # n = 6 and 1 true bit
    >>> twomax(np.array([1, 0, 0, 0, 0, 0]))
    2

    # n = 6 and 1 true bit
    >>> twomax(np.array([0, 0, 1, 0, 0, 0]))
    2

    # n = 6 and 2 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0, 0]))
    3

    # n = 6 and 2 true bits
    >>> twomax(np.array([0, 0, 0, 1, 0, 1]))
    3

    # n = 6 and 2 true bits
    >>> twomax(np.array([1, 0, 0, 0, 1, 0]))
    3

    # n = 6 and 3 true bits
    >>> twomax(np.array([0, 1, 1, 0, 1, 0]))
    4

    # n = 6 and 3 true bits
    >>> twomax(np.array([0, 1, 1, 0, 0, 1]))
    4

    # n = 6 and 3 true bits
    >>> twomax(np.array([1, 0, 1, 0, 1, 0]))
    4

    # n = 6 and 4 true bits
    >>> twomax(np.array([1, 0, 0, 1, 1, 1]))
    3

    # n = 6 and 4 true bits
    >>> twomax(np.array([1, 0, 1, 1, 0, 1]))
    3

    # n = 6 and 4 true bits
    >>> twomax(np.array([1, 0, 0, 1, 1, 1]))
    3

    # n = 6 and 5 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1, 1]))
    2

    # n = 6 and 5 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 1]))
    2

    # n = 6 and 5 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 1]))
    2

    # n = 6 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1]))
    0

    # n = 7 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0]))
    1

    # n = 7 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 7 and 1 true bit
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 0]))
    2

    # n = 7 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 7 and 2 true bits
    >>> twomax(np.array([0, 0, 1, 0, 0, 1, 0]))
    3

    # n = 7 and 2 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0, 0, 0]))
    3

    # n = 7 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 1]))
    3

    # n = 7 and 3 true bits
    >>> twomax(np.array([0, 1, 1, 0, 0, 0, 1]))
    4

    # n = 7 and 3 true bits
    >>> twomax(np.array([1, 0, 0, 0, 1, 1, 0]))
    4

    # n = 7 and 3 true bits
    >>> twomax(np.array([1, 0, 0, 0, 1, 1, 0]))
    4

    # n = 7 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 0, 1, 1, 0]))
    4

    # n = 7 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 1, 0, 0, 1]))
    4

    # n = 7 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 0, 0]))
    4

    # n = 7 and 5 true bits
    >>> twomax(np.array([1, 1, 1, 0, 1, 0, 1]))
    3

    # n = 7 and 5 true bits
    >>> twomax(np.array([1, 1, 1, 0, 0, 1, 1]))
    3

    # n = 7 and 5 true bits
    >>> twomax(np.array([1, 1, 0, 1, 0, 1, 1]))
    3

    # n = 7 and 6 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 1, 1]))
    2

    # n = 7 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 1, 0, 1, 1]))
    2

    # n = 7 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 7 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 8 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    1

    # n = 8 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 1]))
    2

    # n = 8 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 8 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 1, 0, 0, 0]))
    2

    # n = 8 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 1, 0, 0, 0, 0]))
    3

    # n = 8 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 0, 1, 0, 0, 0]))
    3

    # n = 8 and 2 true bits
    >>> twomax(np.array([0, 0, 0, 0, 1, 0, 0, 1]))
    3

    # n = 8 and 3 true bits
    >>> twomax(np.array([0, 0, 0, 0, 1, 0, 1, 1]))
    4

    # n = 8 and 3 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 1, 1, 1]))
    4

    # n = 8 and 3 true bits
    >>> twomax(np.array([0, 0, 0, 1, 1, 1, 0, 0]))
    4

    # n = 8 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 0, 1, 1, 0, 0]))
    5

    # n = 8 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0, 1, 1, 0]))
    5

    # n = 8 and 4 true bits
    >>> twomax(np.array([0, 1, 1, 0, 1, 0, 1, 0]))
    5

    # n = 8 and 5 true bits
    >>> twomax(np.array([1, 1, 0, 1, 0, 1, 1, 0]))
    4

    # n = 8 and 5 true bits
    >>> twomax(np.array([0, 1, 0, 1, 0, 1, 1, 1]))
    4

    # n = 8 and 5 true bits
    >>> twomax(np.array([1, 1, 0, 0, 1, 1, 1, 0]))
    4

    # n = 8 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 1, 0, 1, 1, 0]))
    3

    # n = 8 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 1, 0, 1, 0, 1]))
    3

    # n = 8 and 6 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1, 1, 1, 0]))
    3

    # n = 8 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 8 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 0, 1, 1, 1, 1]))
    2

    # n = 8 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 8 and 8 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 9 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    1

    # n = 9 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0]))
    2

    # n = 9 and 1 true bit
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0]))
    2

    # n = 9 and 1 true bit
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 0, 0, 0]))
    2

    # n = 9 and 2 true bits
    >>> twomax(np.array([0, 0, 0, 1, 0, 0, 0, 1, 0]))
    3

    # n = 9 and 2 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 1, 0, 0, 1]))
    3

    # n = 9 and 2 true bits
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 0, 1, 0]))
    3

    # n = 9 and 3 true bits
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 1, 0, 1]))
    4

    # n = 9 and 3 true bits
    >>> twomax(np.array([0, 1, 0, 1, 0, 0, 0, 0, 1]))
    4

    # n = 9 and 3 true bits
    >>> twomax(np.array([0, 0, 1, 0, 0, 1, 0, 1, 0]))
    4

    # n = 9 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0, 0, 0, 1, 1]))
    5

    # n = 9 and 4 true bits
    >>> twomax(np.array([0, 0, 1, 0, 0, 1, 1, 1, 0]))
    5

    # n = 9 and 4 true bits
    >>> twomax(np.array([0, 1, 0, 1, 0, 1, 0, 0, 1]))
    5

    # n = 9 and 5 true bits
    >>> twomax(np.array([1, 1, 1, 0, 0, 0, 0, 1, 1]))
    5

    # n = 9 and 5 true bits
    >>> twomax(np.array([0, 0, 0, 1, 1, 1, 0, 1, 1]))
    5

    # n = 9 and 5 true bits
    >>> twomax(np.array([0, 0, 1, 1, 0, 1, 1, 1, 0]))
    5

    # n = 9 and 6 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 0, 0, 1, 1]))
    4

    # n = 9 and 6 true bits
    >>> twomax(np.array([1, 0, 1, 1, 0, 1, 1, 0, 1]))
    4

    # n = 9 and 6 true bits
    >>> twomax(np.array([1, 0, 1, 1, 1, 0, 1, 0, 1]))
    4

    # n = 9 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 0, 0, 1, 1]))
    3

    # n = 9 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 0, 0, 1, 1, 1, 1]))
    3

    # n = 9 and 7 true bits
    >>> twomax(np.array([1, 0, 1, 1, 1, 0, 1, 1, 1]))
    3

    # n = 9 and 8 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 1, 1, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 0, 1, 1, 1]))
    2

    # n = 9 and 8 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 0, 1, 1]))
    2

    # n = 9 and 9 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 10 and 0 true bits
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    1

    # n = 10 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 1, 0, 0]))
    2

    # n = 10 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    2

    # n = 10 and 1 true bit
    >>> twomax(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    2

    # n = 10 and 2 true bits
    >>> twomax(np.array([1, 0, 0, 0, 0, 0, 1, 0, 0, 0]))
    3

    # n = 10 and 2 true bits
    >>> twomax(np.array([0, 0, 0, 1, 1, 0, 0, 0, 0, 0]))
    3

    # n = 10 and 2 true bits
    >>> twomax(np.array([0, 0, 1, 0, 0, 0, 0, 0, 1, 0]))
    3

    # n = 10 and 3 true bits
    >>> twomax(np.array([0, 0, 1, 0, 1, 0, 0, 0, 0, 1]))
    4

    # n = 10 and 3 true bits
    >>> twomax(np.array([0, 1, 0, 0, 0, 0, 0, 0, 1, 1]))
    4

    # n = 10 and 3 true bits
    >>> twomax(np.array([0, 1, 1, 0, 1, 0, 0, 0, 0, 0]))
    4

    # n = 10 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 1, 1, 0, 0, 0, 0, 0]))
    5

    # n = 10 and 4 true bits
    >>> twomax(np.array([1, 1, 0, 0, 0, 0, 0, 0, 1, 1]))
    5

    # n = 10 and 4 true bits
    >>> twomax(np.array([1, 0, 1, 0, 0, 1, 1, 0, 0, 0]))
    5

    # n = 10 and 5 true bits
    >>> twomax(np.array([0, 0, 0, 1, 0, 1, 1, 1, 1, 0]))
    6

    # n = 10 and 5 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1, 0, 0, 0, 1, 0]))
    6

    # n = 10 and 5 true bits
    >>> twomax(np.array([1, 0, 0, 1, 1, 1, 0, 1, 0, 0]))
    6

    # n = 10 and 6 true bits
    >>> twomax(np.array([1, 1, 1, 1, 0, 0, 1, 0, 0, 1]))
    5

    # n = 10 and 6 true bits
    >>> twomax(np.array([0, 1, 0, 1, 1, 1, 1, 0, 0, 1]))
    5

    # n = 10 and 6 true bits
    >>> twomax(np.array([0, 1, 0, 1, 1, 0, 1, 1, 1, 0]))
    5

    # n = 10 and 7 true bits
    >>> twomax(np.array([1, 0, 1, 1, 0, 1, 0, 1, 1, 1]))
    4

    # n = 10 and 7 true bits
    >>> twomax(np.array([1, 1, 1, 0, 0, 1, 1, 1, 1, 0]))
    4

    # n = 10 and 7 true bits
    >>> twomax(np.array([0, 1, 0, 1, 0, 1, 1, 1, 1, 1]))
    4

    # n = 10 and 8 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1, 1, 1, 0, 1, 1]))
    3

    # n = 10 and 8 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 0]))
    3

    # n = 10 and 8 true bits
    >>> twomax(np.array([0, 0, 1, 1, 1, 1, 1, 1, 1, 1]))
    3

    # n = 10 and 9 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 1]))
    2

    # n = 10 and 9 true bits
    >>> twomax(np.array([0, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 9 true bits
    >>> twomax(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1, 1]))
    2

    # n = 10 and 10 true bits
    >>> twomax(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0
    """
    s: Final[int] = len(x)
    number_of_ones: Final[int] = int(x.sum())
    return 0 if s == number_of_ones else (
        1 + s - max(number_of_ones, s - number_of_ones))


class TwoMax(BitStringProblem):
    """The TwoMax benchmark problem."""

    def __init__(self, n: int) -> None:
        """
        Initialize the twomax objective function.

        :param n: the dimension of the problem

        >>> TwoMax(2).n
        2
        >>> TwoMax(4).evaluate(np.array([True, True, False, True]))
        2
        """
        super().__init__(n)
        self.evaluate = twomax  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the twomax objective function.

        :return: `twomax_` + length of string

        >>> print(TwoMax(13))
        twomax_13
        """
        return f"twomax_{self.n}"

    def upper_bound(self) -> int:
        """
        Get the upper bound of the twomax problem.

        :return: the length of the bit string integer-divided by 2, plus 1

        >>> TwoMax(15).upper_bound()
        8

        >>> TwoMax(5).upper_bound()
        3
        """
        return (self.n // 2) + 1

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 3, scale_max: int = 333) \
            -> Iterator[Callable[[], "TwoMax"]]:
        """
        Get the 77 default instances of the :class:`TwoMax` problem.

        :param scale_min: the minimum permitted scale, by default `3`
        :param scale_max: the maximum permitted scale, by default `333`
        :returns: a sequence of default :class:`TwoMax` instances

        >>> len(list(TwoMax.default_instances()))
        77

        >>> [x() for x in TwoMax.default_instances()]
        [twomax_3, twomax_4, twomax_5, twomax_6, twomax_7, twomax_8, \
twomax_9, twomax_10, twomax_11, twomax_12, twomax_13, twomax_14, twomax_15, \
twomax_16, twomax_17, twomax_18, twomax_19, twomax_20, twomax_21, twomax_22, \
twomax_23, twomax_24, twomax_25, twomax_26, twomax_27, twomax_28, twomax_29, \
twomax_30, twomax_31, twomax_32, twomax_33, twomax_36, twomax_40, twomax_41, \
twomax_42, twomax_44, twomax_48, twomax_49, twomax_50, twomax_55, twomax_59, \
twomax_60, twomax_64, twomax_66, twomax_70, twomax_77, twomax_79, twomax_80, \
twomax_81, twomax_85, twomax_88, twomax_90, twomax_96, twomax_99, \
twomax_100, twomax_107, twomax_111, twomax_121, twomax_125, twomax_128, \
twomax_144, twomax_149, twomax_169, twomax_170, twomax_192, twomax_196, \
twomax_199, twomax_200, twomax_222, twomax_225, twomax_243, twomax_256, \
twomax_269, twomax_289, twomax_300, twomax_324, twomax_333]
        """
        return cast("Iterator[Callable[[], TwoMax]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
Parsers for structured log data produced by the `moptipy` experiment API.

The `moptipy` :class:`~moptipy.api.execution.Execution` and experiment-running
facility (:func:`~moptipy.api.experiment.run_experiment`) uses the class
:class:`~moptipy.utils.logger.Logger` from module :mod:`~moptipy.utils.logger`
to produce log files complying with
https://thomasweise.github.io/moptipy/#log-files.

Here we provide a skeleton for parsing such log files in form of the class
:class:`~LogParser`. It works similar to SAX-XML parsing in that the data
is read is from files and methods that consume the data are invoked. By
overwriting these methods, we can do useful things with the data.

For example in module :mod:`~moptipy.evaluation.end_results`, the method
:meth:`~moptipy.evaluation.end_results.from_logs` can load
:class:`~moptipy.evaluation.end_results.EndResult` records from the logs
and the method :meth:`~moptipy.evaluation.progress.from_logs` in
module :mod:`~moptipy.evaluation.progress` reads the whole
:class:`~moptipy.evaluation.progress.Progress` that the algorithms make
over time.
"""

from contextlib import suppress
from math import inf, isfinite, isinf
from typing import Callable, Final, TypeVar

from pycommons.io.csv import COMMENT_START, SCOPE_SEPARATOR
from pycommons.io.parser import Parser
from pycommons.io.path import Path
from pycommons.strings.string_conv import str_to_num
from pycommons.types import check_to_int_range, type_error

from moptipy.api.logging import (
    ERROR_SECTION_PREFIX,
    FILE_SUFFIX,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_NAME,
    KEY_RAND_SEED,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
    SCOPE_ALGORITHM,
    SCOPE_ENCODING,
    SCOPE_OBJECTIVE_FUNCTION,
    SCOPE_PROCESS,
    SECTION_FINAL_STATE,
    SECTION_SETUP,
)
from moptipy.evaluation._utils import _check_max_time_millis
from moptipy.utils.logger import (
    SECTION_END,
    SECTION_START,
    parse_key_values,
)
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.strings import (
    PART_SEPARATOR,
    sanitize_name,
)

#: the maximum FEs of a black-box process
_FULL_KEY_MAX_FES: Final[str] = \
    f"{SCOPE_PROCESS}{SCOPE_SEPARATOR}{KEY_MAX_FES}"
#: the maximum runtime in milliseconds of a black-box process
_FULL_KEY_MAX_TIME_MILLIS: Final[str] = \
    f"{SCOPE_PROCESS}{SCOPE_SEPARATOR}{KEY_MAX_TIME_MILLIS}"
#: the goal objective value of a black-box process
_FULL_KEY_GOAL_F: Final[str] = f"{SCOPE_PROCESS}{SCOPE_SEPARATOR}{KEY_GOAL_F}"
#: the random seed
_FULL_KEY_RAND_SEED: Final[str] = \
    f"{SCOPE_PROCESS}{SCOPE_SEPARATOR}{KEY_RAND_SEED}"
#: the full algorithm name key
_FULL_KEY_ALGORITHM: Final[str] = \
    f"{SCOPE_ALGORITHM}{SCOPE_SEPARATOR}{KEY_NAME}"
#: the full objective function name key
_FULL_KEY_OBJECTIVE: Final[str] = \
    f"{SCOPE_OBJECTIVE_FUNCTION}{SCOPE_SEPARATOR}{KEY_NAME}"
#: the full encoding name key
_FULL_KEY_ENCODING: Final[str] = \
    f"{SCOPE_ENCODING}{SCOPE_SEPARATOR}{KEY_NAME}"


def _true(_) -> bool:
    """
    Get `True` as return value, always.

    :retval `True`: always
    """
    return True


#: the type variable for data to be read from the directories
T = TypeVar("T")


class LogParser[T](Parser[T]):
    """
    A log parser can parse a log file and separate the sections.

    The log parser is designed to load data from text files generated
    by :class:`~moptipy.utils.logger.FileLogger`. It can also recursively
    parse directories.
    """

    def __init__(self, path_filter: Callable[[Path], bool] | None = None):
        """
        Initialize the log parser.

        :param path_filter: a filter allowing us to skip paths or files. If
            this :class:`Callable` returns `True`, the file or directory is
            considered for parsing. If it returns `False`, it is skipped.
        """
        if path_filter is None:
            path_filter = _true
        elif not callable(path_filter):
            raise type_error(path_filter, "path_filter", call=True)
        #: the current depth in terms of directories
        self.__depth: int = 0
        #: the path filter
        self.__path_filter: Final[Callable[[Path], bool]] = path_filter

    def _should_list_directory(self, directory: Path) -> tuple[bool, bool]:
        """
        Decide whether to enter a directory to parse all files inside.

        :param directory: the path of the directory
        :return: a tuple with two `True` values if all the sub-directories and
            files inside the directory should be processed, two `False` values
            if this directory should be skipped and parsing should continue
            with the next sibling directory
        """
        should: Final[bool] = self.__path_filter(directory)
        return should, should

    def _should_parse_file(self, file: Path) -> bool:
        """
        Decide whether to start parsing a file.

        :param file: the file path
        :return: `True` if the file should be parsed, `False` if it should be
            skipped (and
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` should
            return `True`).
        """
        return file.endswith(FILE_SUFFIX) and self.__path_filter(file)

    # noinspection PyMethodMayBeStatic
    def _start_section(self, title: str) -> bool:
        """
        Start a section.

        If this method returns `True`, then all the lines of text of the
        section `title` will be read and together passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.lines`.
        If this method returns `False`, then the section will be skipped
        and we fast-forward to the next section, if any.

        :param title: the section title
        :return: `True` if the section data should be loaded and passed to
            :meth:`lines`, `False` of the section can be skipped. In that
            case, we will fast-forward to the next
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`.
        """
        if not title:
            raise ValueError(f"Title cannot be empty, but is {title!r}.")
        if title.startswith(ERROR_SECTION_PREFIX):
            raise ValueError(f"Encountered error section {title!r}.")
        return False

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def _lines(self, lines: list[str]) -> bool:
        """
        Consume all the lines from a section.

        This method receives the complete text of a section, where all lines
        are separated and put into one list `lines`. Each line is stripped
        from whitespace and comments, empty lines are omitted.
        If this method returns `True`, we will continue parsing the file and
        move to the next section, if any, or directly to the end of the file
        parsing process.

        :param lines: the lines to consume
        :return: `True` if further parsing is necessary and the next section
            should be fed to
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`,
            `False` if the parsing process can be terminated`
        """
        del lines
        return True

    def _parse_file(self, file: Path) -> T | None:  # pylint: disable=R1711
        """
        Parse the contents of a file.

        :param file: the file to parse
        :return: the return value received from invoking `get_result`
        """
        lines: list[str] = []
        buffer: list[str] = []
        state: int = 0
        wants_section: bool = False
        sec_end: str = ""
        section: str = ""
        sect_start: Final[str] = SECTION_START
        sect_end: Final[str] = SECTION_END
        cmt_chr: Final[str] = COMMENT_START

        index: int = 0
        with (file.open_for_read() as handle):
            while True:

                # get the next line
                if index >= len(buffer):
                    try:
                        buffer = handle.readlines(128)
                    except Exception as be:
                        raise ValueError(
                            f"Error when reading lines from file {file!r} "
                            f"while in section {section!r}."
                            if state == 1 else
                            "Error when reading lines from file "
                            f"{file!r}.") from be
                    if (buffer is None) or (len(buffer) <= 0):
                        break
                    index = 0

                orig_cur = buffer[index]
                index += 1

                # strip next line from comments and white space
                cur = orig_cur.strip()
                if len(cur) <= 0:
                    continue

                i = cur.find(cmt_chr)
                if i >= 0:
                    cur = cur[:i].strip()
                    if len(cur) <= 0:
                        continue

                if state in {0, 2}:
                    if not cur.startswith(sect_start):
                        raise ValueError("Line should start with "
                                         f"{sect_start!r} but is "
                                         f"{orig_cur!r} in file {file!r}.")
                    section = cur[len(sect_start):]
                    if len(section) <= 0:
                        raise ValueError(
                            "Section title cannot be empty in "
                            f"{file!r}, but encountered {orig_cur!r}.")
                    state = 1
                    sec_end = sect_end + section
                    wants_section = self._start_section(section)
                elif state == 1:
                    if cur == sec_end:
                        state = 2
                        if wants_section:
                            try:
                                do_next = self._lines(lines)
                            except Exception as be:
                                raise ValueError(
                                    "Error when processing section "
                                    f"{section!r} in file {file!r}.") \
                                    from be
                            lines.clear()
                            if not do_next:
                                break
                    elif wants_section:
                        lines.append(cur)

        if state == 0:
            raise ValueError(f"Log file {file!r} contains no section.")
        if state == 1:
            raise ValueError(f"Log file {file!r} ended before"
                             f"encountering {sec_end!r}.")
        return None  # pylint: disable=R1711


#: the start for random seeds
_SEED_START: Final[str] = f"{PART_SEPARATOR}0x"


class ExperimentParser[T](LogParser[T]):
    """A log parser following our pre-defined experiment structure."""

    def __init__(self, path_filter: Callable[[Path], bool] | None = None):
        """
        Initialize the experiment parser.

        :param path_filter: a filter allowing us to skip paths or files. If
            this :class:`Callable` returns `True`, the file or directory is
            considered for parsing. If it returns `False`, it is skipped.
        """
        super().__init__(path_filter)

        #: The name of the algorithm to which the current log file belongs.
        self.algorithm: str | None = None
        #: The name of the instance to which the current log file belongs.
        self.instance: str | None = None
        #: The random seed of the current log file.
        self.rand_seed: int | None = None
        #: the file basename
        self.__file_base_name: str | None = None

    def _start_parse_file(self, file: Path) -> None:
        """
        Start parsing the file.

        This function sets up best guesses about the instance name, the
        algorithm name, and the random seed based on the file name.

        :param file: the file to parse
        """
        super()._start_parse_file(file)
        inst_name_suggestion: str | None = None
        algo_name_suggestion: str | None = None
        with (suppress(Exception)):
            inst_dir: Final[Path] = file.up()
            inst_name_suggestion = inst_dir.basename()
            if sanitize_name(inst_name_suggestion) != inst_name_suggestion:
                inst_name_suggestion = None
            else:
                algo_dir: Final[Path] = inst_dir.up()
                algo_name_suggestion = algo_dir.basename()
                if sanitize_name(algo_name_suggestion) \
                        != algo_name_suggestion:
                    algo_name_suggestion = None

        fbn: Final[str] = file.basename()
        self.__file_base_name = fbn

        seed_start: int = fbn.rfind(_SEED_START)
        seed_end: int = str.__len__(fbn) - len(FILE_SUFFIX)
        if (seed_start > 0) and (seed_end > (seed_start + 3)):
            try:
                self.rand_seed = rand_seed_check(int(
                    fbn[seed_start + 3:seed_end], base=16))
            except Exception:  # noqa
                seed_start = -1
        if (seed_start > 0) and (inst_name_suggestion is not None):
            if algo_name_suggestion is not None:
                start: str = (f"{algo_name_suggestion}{PART_SEPARATOR}"
                              f"{inst_name_suggestion}{_SEED_START}")
                if fbn.casefold().startswith(start.casefold()):
                    self.instance = inst_name_suggestion
                    self.algorithm = algo_name_suggestion
            else:
                start = f"{PART_SEPARATOR}{inst_name_suggestion}{_SEED_START}"
                if start.casefold() in fbn.casefold():
                    self.instance = inst_name_suggestion

    def _parse_file(self, file: Path) -> T | None:
        """
        Parse the file contents.

        :param file: the file to parse
        :returns: nothing
        """
        res: Final[T | None] = super()._parse_file(file)

        if (self.algorithm is not None) and (self.rand_seed is not None) and (
                self.instance is None):
            bn: Final[str] = self.__file_base_name
            alcf: str = f"{self.algorithm.casefold()}{PART_SEPARATOR}"
            if bn.casefold().startswith(alcf):
                inst_end: Final[int] = bn.rfind(_SEED_START)
                anl: Final[int] = str.__len__(alcf)
                if inst_end > anl:
                    inst: str = bn[anl:inst_end]
                    if sanitize_name(inst) == inst:
                        self.instance = inst
            self.instance = "unknown"

        return res

    def _end_parse_file(self, file: Path) -> None:
        """
        Finalize parsing a file.

        :param file: the file
        """
        self.rand_seed = None
        self.algorithm = None
        self.instance = None
        self.__file_base_name = None
        super()._end_parse_file(file)


class SetupAndStateParser[T](ExperimentParser[T]):
    """
    A log parser which loads and processes the basic data from the logs.

    This parser processes the `SETUP` and `STATE` sections of a log file and
    stores the performance-related information in member variables.
    """

    def __init__(self, path_filter: Callable[[Path], bool] | None = None):
        """
        Create the basic data parser.

        :param path_filter: a filter allowing us to skip paths or files. If
            this :class:`Callable` returns `True`, the file or directory is
            considered for parsing. If it returns `False`, it is skipped.
        """
        super().__init__(path_filter)
        #: the total consumed runtime, in objective function evaluations
        self.total_fes: int | None = None
        #: the total consumed runtime in milliseconds
        self.total_time_millis: int | None = None
        #: the best objective function value encountered
        self.best_f: int | float | None = None
        #: the objective function evaluation when the last improvement
        #: happened, in milliseconds
        self.last_improvement_fe: int | None = None
        #: the time step when the last improvement happened, in milliseconds
        self.last_improvement_time_millis: int | None = None
        #: the goal objective value, if any
        self.goal_f: int | float | None = None
        #: the maximum permitted number of objective function evaluations,
        #: if any
        self.max_fes: int | None = None
        #: the maximum runtime limit in milliseconds, if any
        self.max_time_millis: int | None = None
        #: The name of the objective to which the current log file belongs.
        self.objective: str | None = None
        #: The name of the encoding to which the current log file belongs.
        self.encoding: str | None = None
        #: the internal state, an OR mask: 1=after setup section, 2=after
        #: state section, 4=in setup section, 8=in state section
        self.__state: int = 0

    def _should_parse_file(self, file: Path) -> bool:
        """
        Begin parsing the file identified by `path`.

        :param file: the path identifying the file
        """
        if not super()._should_parse_file(file):
            return False
        if self.__state != 0:
            raise ValueError(f"Illegal state when trying to parse {file}.")
        return True

    def _parse_file(self, file: Path) -> T | None:
        """
        Parse the file.

        :param file: the file
        :returns: the parsed object
        """
        res: Final[T | None] = super()._parse_file(file)
        if self.__state != 3:
            raise ValueError(
                "Illegal state, log file must have both a "
                f"{SECTION_FINAL_STATE!r} and a "
                f"{SECTION_SETUP!r} section.")
        if self.rand_seed is None:
            raise ValueError("rand_seed is missing.")
        if self.algorithm is None:
            raise ValueError("algorithm is missing.")
        if self.instance is None:
            raise ValueError("instance is missing.")
        if self.objective is None:
            raise ValueError("objective is missing.")
        if self.total_fes is None:
            raise ValueError("total_fes is missing.")
        if self.total_time_millis is None:
            raise ValueError("total_time_millis is missing.")
        if self.best_f is None:
            raise ValueError("best_f is missing.")
        if self.last_improvement_fe is None:
            raise ValueError("last_improvement_fe is missing.")
        if self.last_improvement_time_millis is None:
            raise ValueError("last_improvement_time_millis is missing.")
        return res

    def _end_parse_file(self, file: Path) -> None:
        """
        Finalize the state *after* parsing.

        :param file: the file to parse
        """
        self.total_fes = None
        self.total_time_millis = None
        self.best_f = None
        self.last_improvement_fe = None
        self.last_improvement_time_millis = None
        self.goal_f = None
        self.max_fes = None
        self.max_time_millis = None
        self.objective = None
        self.encoding = None
        self.__state = 0
        return super()._end_parse_file(file)

    def _needs_more_lines(self) -> bool:
        """
        Check whether we need to process more lines.

        You can overwrite this method if your parser parses additional log
        sections. Your overwritten method should return `True` if more
        sections except `STATE` and `SETUP` still need to be parsed and return
        `super().needs_more_lines()` otherwise.

        :returns: `True` if more data needs to be processed, `False` otherwise
        """
        return self.__state != 3

    def _lines(self, lines: list[str]) -> bool:
        """
        Process the lines loaded from a section.

        If you process more sections, you should override this method. Your
        overridden method then can parse the data if you are in the right
        section. It should end with `return super().lines(lines)`.

        :param lines: the lines that have been loaded
        :returns: `True` if parsing should be continued, `False` otherwise
        """
        if (self.__state & 4) != 0:
            self._setup_section(parse_key_values(lines))
        elif (self.__state & 8) != 0:
            self._state_section(lines)
        return self._needs_more_lines()

    def _start_section(self, title: str) -> bool:
        """
        Begin a section.

        :param title: the section title
        :returns: `True` if the text of the section should be processed,
            `False` otherwise
        """
        super()._start_section(title)
        if title == SECTION_SETUP:
            if (self.__state & 1) != 0:
                raise ValueError(f"Already did section {title!r}.")
            self.__state |= 4
            return True
        if title == SECTION_FINAL_STATE:
            if (self.__state & 2) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 8
            return True
        return False

    def _setup_section(self, data: dict[str, str]) -> None:
        """
        Parse the data from the `setup` section.

        :param data: the parsed data
        """
        self.goal_f = None
        if _FULL_KEY_GOAL_F in data:
            goal_f = data[_FULL_KEY_GOAL_F]
            g: Final[int | float] = str_to_num(goal_f)
            if isfinite(g):
                self.goal_f = g
            elif not (isinf(g) and (g >= inf)):
                raise ValueError(
                    f"invalid goal f {goal_f}, which renders to {g}")

        if _FULL_KEY_MAX_FES in data:
            self.max_fes = check_to_int_range(
                data[_FULL_KEY_MAX_FES], _FULL_KEY_MAX_FES, 1,
                1_000_000_000_000_000)
        if _FULL_KEY_MAX_TIME_MILLIS in data:
            self.max_time_millis = check_to_int_range(
                data[_FULL_KEY_MAX_TIME_MILLIS], _FULL_KEY_MAX_TIME_MILLIS, 1,
                1_000_000_000_000)
        if _FULL_KEY_ALGORITHM in data:
            a = data[_FULL_KEY_ALGORITHM]
            if self.algorithm is None:
                self.algorithm = a
            elif a != self.algorithm:
                # this error may occur under windows due to case-insensitive
                # file names
                if a.casefold() == self.algorithm.casefold():
                    self.algorithm = a  # rely on name from log file
                else:  # ok, case was not the issue - raise error
                    raise ValueError(
                        f"algorithm name from file name is {self.algorithm!r}"
                        f", but key {_FULL_KEY_ALGORITHM!r} gives {a!r}.")
        else:
            raise ValueError(f"key {_FULL_KEY_ALGORITHM!r} missing in file!")

        if _FULL_KEY_OBJECTIVE in data:
            self.objective = data[_FULL_KEY_OBJECTIVE]
        else:
            raise ValueError(f"key {_FULL_KEY_OBJECTIVE!r} missing in file!")

        self.encoding = data.get(_FULL_KEY_ENCODING)

        seed_check = rand_seed_check(int(data[_FULL_KEY_RAND_SEED]))
        if self.rand_seed is None:
            self.rand_seed = seed_check
        elif seed_check != self.rand_seed:
            raise ValueError(
                f"Found seed {seed_check} in log file, but file name "
                f"indicates seed {self.rand_seed}.")

        self.__state = (self.__state | 1) & (~4)

    def _state_section(self, lines: list[str]) -> None:
        """
        Process the data of the final state section.

        :param lines: the lines of that section
        """
        data: Final[dict[str, str]] = parse_key_values(lines)

        self.total_fes = check_to_int_range(
            data[KEY_TOTAL_FES], KEY_TOTAL_FES, 1,
            1_000_000_000_000_000 if self.max_fes is None else self.max_fes)
        self.total_time_millis = check_to_int_range(
            data[KEY_TOTAL_TIME_MILLIS], KEY_TOTAL_TIME_MILLIS, 0,
            1_000_000_000_000 if self.max_time_millis is None else
            ((1_000_000 + self.max_time_millis) * 1_000))
        if self.max_time_millis is not None:
            _check_max_time_millis(self.max_time_millis, self.total_fes,
                                   self.total_time_millis)
        self.best_f = str_to_num(data[KEY_BEST_F])
        if not isfinite(self.best_f):
            raise ValueError(f"infinite best f detected: {self.best_f}")
        self.last_improvement_fe = check_to_int_range(
            data[KEY_LAST_IMPROVEMENT_FE], KEY_LAST_IMPROVEMENT_FE, 1,
            self.total_fes)
        self.last_improvement_time_millis = check_to_int_range(
            data[KEY_LAST_IMPROVEMENT_TIME_MILLIS],
            KEY_LAST_IMPROVEMENT_TIME_MILLIS, 0, self.total_time_millis)
        self.__state = (self.__state | 2) & (~8)



"""
A multi-normal distribution.

>>> from moptipy.utils.nputils import rand_generator
>>> gen = rand_generator(12)

>>> from moptipy.operators.intspace.op0_random import Op0Random
>>> space = IntSpace(4, -2, 2)
>>> op0 = Op0Random(space)
>>> x1 = space.create()
>>> op0.op0(gen, x1)
>>> x1
array([-1,  2,  2,  1], dtype=int8)

>>> op1 = Op1MNormal(space, 1, True, 1.5)
>>> op1.initialize()
>>> x2 = space.create()
>>> op1.op1(gen, x2, x1)
>>> x2
array([-1,  0,  0,  1], dtype=int8)
>>> op1.op1(gen, x2, x1)
>>> x2
array([-1,  2,  1, -2], dtype=int8)
>>> op1.op1(gen, x2, x1)
>>> x2
array([-1,  2,  2,  0], dtype=int8)

>>> space = IntSpace(12, 0, 20)
>>> op0 = Op0Random(space)
>>> x1 = space.create()
>>> op0.op0(gen, x1)
>>> x1
array([ 6,  1, 18, 11, 13, 11,  2,  3, 13,  7, 10, 14], dtype=int8)

>>> op1 = Op1MNormal(space, 1, True, 1.5)
>>> op1.initialize()
>>> x2 = space.create()
>>> op1.op1(gen, x2, x1)
>>> x2
array([ 6,  1, 18, 11, 13, 11,  2,  3, 12,  7,  9, 13], dtype=int8)
>>> op1.op1(gen, x2, x1)
>>> x2
array([ 6,  1, 20, 11, 13, 11,  2,  3, 13,  7, 10, 14], dtype=int8)
>>> op1.op1(gen, x2, x1)
>>> x2
array([ 3,  1, 18, 11, 13, 11,  2,  3, 13,  7,  9, 14], dtype=int8)

>>> str(op1)
'normB1_1d5'
"""
from math import ceil, floor, isfinite
from typing import Final

import numba  # type: ignore
import numpy as np
from numpy.random import Generator
from pycommons.types import check_int_range, type_error

from moptipy.api.operators import Op1
from moptipy.spaces.intspace import IntSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import (
    fill_in_canonical_permutation,
    int_range_to_dtype,
)
from moptipy.utils.strings import num_to_str_for_name


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def _mnormal(m: int, none_is_ok: bool, permutation: np.ndarray,
             random: Generator, sd: float, min_val: int, max_val: int,
             dest: np.ndarray, x: np.ndarray) -> None:
    """
    Copy `x` into `dest` and sample normal distribution for each element n/m.

    This method will first copy `x` to `dest`. Then it will decide for each
    value in `dest` whether it should be changed: The change happens with
    probability `m/n`, where `n` is the length of `dest`.
    Regardless of the probability, at least one element will always be
    changed if self.at_least_1 is True.

    :param m: the value of m
    :param none_is_ok: is it OK to flip nothing?
    :param permutation: the internal permutation
    :param random: the random number generator
    :param sd: the standard deviation to be used for the normal distribution
    :param min_val: the minimal permissible value
    :param max_val: the maximal permissible value
    :param dest: the destination array to receive the new point
    :param x: the existing point in the search space
    """
    dest[:] = x[:]  # copy source to destination
    length: Final[int] = len(dest)  # get n
    p: Final[float] = m / length  # probability to change values

    flips: int  # the number of values to flip
    while True:
        flips = random.binomial(length, p)  # get number of values to change
        if flips > 0:
            break  # we will change some values
        if none_is_ok:
            return  # we will change no values

    i: int = length
    end: Final[int] = length - flips
    while i > end:  # we iterate from i=length down to end=length-change
        k = random.integers(0, i)  # index of next value index in permutation
        i -= 1  # decrease i
        idx = permutation[k]  # get index of bit to value and move to end
        permutation[i], permutation[k] = idx, permutation[i]

        # put a normal distribution around old value and sample
        # a new value
        old_value = dest[idx]
        while True:
            rnd = random.normal(scale=sd)
            new_value = old_value + (ceil(rnd) if rnd > 0 else floor(rnd))
            if (new_value != old_value) and (min_val <= new_value <= max_val):
                break
        dest[idx] = new_value


class Op1MNormal(Op1):
    """Randomly choose a number of ints to change with normal distribution."""

    def __init__(self, space: IntSpace, m: int = 1, at_least_1: bool = True,
                 sd: float = 1.0):
        """
        Initialize the operator.

        :param n: the length of the bit strings
        :param m: the factor for computing the probability of flipping
            the bits
        :param at_least_1: should at least one bit be flipped?
        """
        super().__init__()
        if not isinstance(space, IntSpace):
            raise type_error(space, "space", IntSpace)
        #: the internal dimension
        self.__n: Final[int] = space.dimension
        #: the minimum permissible value
        self.__min: Final[int] = space.min_value
        #: the maximum permissible value
        self.__max: Final[int] = space.max_value
        #: the value of m in p=m/n
        self.__m: Final[int] = check_int_range(m, "m", 1, self.__n)
        if not isinstance(at_least_1, bool):
            raise type_error(at_least_1, "at_least_1", bool)
        #: is it OK to not flip any bit?
        self.__none_is_ok: Final[bool] = not at_least_1
        #: the internal permutation
        self.__permutation: Final[np.ndarray] = np.empty(
            self.__n, dtype=int_range_to_dtype(0, self.__n - 1))
        if not isinstance(sd, float):
            raise type_error(sd, "sd", float)
        if not (isfinite(sd) and (0 < sd <= (self.__max - self.__min + 1))):
            raise ValueError(
                f"Invalid value {sd} for sd with {self.__min}..{self.__max}.")
        #: the internal standard deviation
        self.__sd: Final[float] = sd

    def initialize(self) -> None:
        """Initialize this operator."""
        super().initialize()
        fill_in_canonical_permutation(self.__permutation)

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray) -> None:
        """
        Copy `x` into `dest` and change each value with probability m/n.

        This method will first copy `x` to `dest`. Then it will change each
        value in `dest` with probability `m/n`, where `n` is the length of
        `dest`. Regardless of the probability, at least one value will always
        be changed if self.at_least_1 is True.

        If a value is changed, we will put a normal distribution around it and
        sample it from that. Of course, we only accept values within the
        limits of the integer space and that are different from the original
        value.

        :param self: the self pointer
        :param random: the random number generator
        :param dest: the destination array to receive the new point
        :param x: the existing point in the search space
        """
        _mnormal(self.__m, self.__none_is_ok, self.__permutation, random,
                 self.__sd, self.__min, self.__max, dest, x)

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :return: "fileB" + m + "n" if none-is-ok else ""
        """
        return (f"normB{self.__m}{'n' if self.__none_is_ok else ''}_"
                f"{num_to_str_for_name(self.__sd)}")

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this operator to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("m", self.__m)
        logger.key_value("n", self.__n)
        logger.key_value("sd", self.__sd)
        logger.key_value("min", self.__min)
        logger.key_value("max", self.__max)



"""
An objective function counting the leading ones in a bit string.

LeadingOnes is a standard benchmark problem in evolutionary computation.
It attempts to maximize the number of consecutive True bits at the beginning
of a bit string. Its best possible objective value is 0, the worst possible
is `n`.

1. L. Darrell Whitley. The GENITOR Algorithm and Selection Pressure: Why
   Rank-Based Allocation of Reproductive Trials is Best. In J. David Schaffer,
   ed., Proceedings of the 3rd International Conference on Genetic Algorithms
   (ICGA'89), June 4-7, 1989, Fairfax, VA, USA, pages 116-121. San Francisco,
   CA, USA: Morgan Kaufmann Publishers Inc. ISBN: 1-55860-066-3
   https://www.researchgate.net/publication/2527551
2. Günter Rudolph, Convergence Properties of Evolutionary Algorithms. Hamburg,
   Germany: Verlag Dr. Kovač, 1997.
3. Peyman Afshani, Manindra Agrawal, Benjamin Doerr, Kasper Green Larsen,
   Kurt Mehlhorn, and Carola Winzen. The Query Complexity of Finding a Hidden
   Permutation. Space-Efficient Data Structures, Streams, and Algorithms.
   pp. 1-11. Springer. 2013.
   doi: https://doi.org/10.1007/978-3-642-40273-9_1
4. Peyman Afshani, Manindra Agrawal, Benjamin Doerr, Carola Doerr, Kasper
   Green Larsen, Kurt Mehlhorn. The Query Complexity of a Permutation-Based
   Variant of Mastermind. Discrete Applied Mathematics. 260:28-50. 2019.
   doi: https://doi.org/10.1016/j.dam.2019.01.007
5. Stefan Droste, Thomas Jansen, and Ingo Wegener. On the Analysis of the
   (1+1) Evolutionary Algorithm. Theoretical Computer Science.
   276(1-2):51-81. 2002.
   doi: https://doi.org/10.1016/S0304-3975(01)00182-7.
6. Denis Antipov, Benjamin Doerr, and Vitalii Karavaev. A Tight Runtime
   Analysis for the (1+(λ,λ)) GA on LeadingOnes. FOGA 2019, pp. 169-182. ACM.
   doi: https://doi.org/10.1145/3299904.3340317.
7. Vitalii Karavaev, Denis Antipov, and Benjamin Doerr. Theoretical and
   Empirical Study of the (1+(λ,λ)) EA on the LeadingOnes Problem. GECCO
   (Companion) 2019. pp 2036-2039. ACM.
   doi: https://doi.org/10.1145/3319619.3326910.
8. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
9. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698
"""
from typing import Callable, Final, Iterator, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def leadingones(x: np.ndarray) -> int:
    """
    Get the length of the string minus the number of leading ones.

    :param x: the np array
    :return: the number of leading ones

    >>> leadingones(np.array([False, False, True, False, False]))
    5
    >>> leadingones(np.array([True, False, False, True, True]))
    4
    >>> leadingones(np.array([True, True, False, False, False]))
    3
    >>> leadingones(np.array([True, True, True, False, True]))
    2
    >>> leadingones(np.array([True, True, True, True, False]))
    1
    >>> leadingones(np.array([True, True, True, True, True]))
    0

    # n = 1 and 0 true bits
    >>> leadingones(np.array([0]))
    1

    # n = 1 and 1 true bit
    >>> leadingones(np.array([1]))
    0

    # n = 2 and 0 true bits
    >>> leadingones(np.array([0, 0]))
    2

    # n = 2 and 1 true bit
    >>> leadingones(np.array([1, 0]))
    1

    # n = 2 and 1 true bit
    >>> leadingones(np.array([0, 1]))
    2

    # n = 2 and 1 true bit
    >>> leadingones(np.array([1, 0]))
    1

    # n = 2 and 2 true bits
    >>> leadingones(np.array([1, 1]))
    0

    # n = 3 and 0 true bits
    >>> leadingones(np.array([0, 0, 0]))
    3

    # n = 3 and 1 true bit
    >>> leadingones(np.array([0, 1, 0]))
    3

    # n = 3 and 1 true bit
    >>> leadingones(np.array([0, 0, 1]))
    3

    # n = 3 and 1 true bit
    >>> leadingones(np.array([0, 1, 0]))
    3

    # n = 3 and 2 true bits
    >>> leadingones(np.array([1, 1, 0]))
    1

    # n = 3 and 2 true bits
    >>> leadingones(np.array([1, 0, 1]))
    2

    # n = 3 and 2 true bits
    >>> leadingones(np.array([1, 1, 0]))
    1

    # n = 3 and 3 true bits
    >>> leadingones(np.array([1, 1, 1]))
    0

    # n = 4 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0]))
    4

    # n = 4 and 1 true bit
    >>> leadingones(np.array([1, 0, 0, 0]))
    3

    # n = 4 and 1 true bit
    >>> leadingones(np.array([0, 0, 1, 0]))
    4

    # n = 4 and 1 true bit
    >>> leadingones(np.array([1, 0, 0, 0]))
    3

    # n = 4 and 2 true bits
    >>> leadingones(np.array([1, 0, 0, 1]))
    3

    # n = 4 and 2 true bits
    >>> leadingones(np.array([0, 1, 0, 1]))
    4

    # n = 4 and 2 true bits
    >>> leadingones(np.array([0, 1, 1, 0]))
    4

    # n = 4 and 3 true bits
    >>> leadingones(np.array([1, 1, 1, 0]))
    1

    # n = 4 and 3 true bits
    >>> leadingones(np.array([1, 1, 1, 0]))
    1

    # n = 4 and 3 true bits
    >>> leadingones(np.array([0, 1, 1, 1]))
    4

    # n = 4 and 4 true bits
    >>> leadingones(np.array([1, 1, 1, 1]))
    0

    # n = 5 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0]))
    5

    # n = 5 and 1 true bit
    >>> leadingones(np.array([0, 1, 0, 0, 0]))
    5

    # n = 5 and 1 true bit
    >>> leadingones(np.array([0, 1, 0, 0, 0]))
    5

    # n = 5 and 1 true bit
    >>> leadingones(np.array([0, 1, 0, 0, 0]))
    5

    # n = 5 and 2 true bits
    >>> leadingones(np.array([1, 1, 0, 0, 0]))
    3

    # n = 5 and 2 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 0]))
    4

    # n = 5 and 2 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 0]))
    5

    # n = 5 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 1]))
    4

    # n = 5 and 3 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 0]))
    2

    # n = 5 and 3 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 1]))
    5

    # n = 5 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1]))
    3

    # n = 5 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1]))
    3

    # n = 5 and 4 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0]))
    1

    # n = 5 and 5 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1]))
    0

    # n = 6 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0]))
    6

    # n = 6 and 1 true bit
    >>> leadingones(np.array([0, 0, 1, 0, 0, 0]))
    6

    # n = 6 and 1 true bit
    >>> leadingones(np.array([0, 1, 0, 0, 0, 0]))
    6

    # n = 6 and 1 true bit
    >>> leadingones(np.array([1, 0, 0, 0, 0, 0]))
    5

    # n = 6 and 2 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 0, 1]))
    6

    # n = 6 and 2 true bits
    >>> leadingones(np.array([0, 0, 1, 0, 0, 1]))
    6

    # n = 6 and 2 true bits
    >>> leadingones(np.array([0, 0, 1, 0, 1, 0]))
    6

    # n = 6 and 3 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 1, 0]))
    6

    # n = 6 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 1, 1]))
    5

    # n = 6 and 3 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 1, 0]))
    6

    # n = 6 and 4 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 1, 1]))
    6

    # n = 6 and 4 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 0]))
    6

    # n = 6 and 4 true bits
    >>> leadingones(np.array([0, 1, 1, 0, 1, 1]))
    6

    # n = 6 and 5 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1]))
    5

    # n = 6 and 5 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 0]))
    1

    # n = 6 and 5 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1]))
    5

    # n = 6 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 1]))
    0

    # n = 7 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0]))
    7

    # n = 7 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 1]))
    7

    # n = 7 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 1, 0, 0]))
    7

    # n = 7 and 1 true bit
    >>> leadingones(np.array([0, 0, 1, 0, 0, 0, 0]))
    7

    # n = 7 and 2 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 0, 1, 0]))
    7

    # n = 7 and 2 true bits
    >>> leadingones(np.array([0, 0, 1, 0, 0, 1, 0]))
    7

    # n = 7 and 2 true bits
    >>> leadingones(np.array([0, 0, 0, 1, 0, 1, 0]))
    7

    # n = 7 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 0, 1, 1]))
    6

    # n = 7 and 3 true bits
    >>> leadingones(np.array([1, 0, 1, 0, 0, 1, 0]))
    6

    # n = 7 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 1, 0, 0]))
    6

    # n = 7 and 4 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 1, 1, 0]))
    6

    # n = 7 and 4 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 0, 1, 0]))
    6

    # n = 7 and 4 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 0, 0]))
    4

    # n = 7 and 5 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 0, 1, 1]))
    7

    # n = 7 and 5 true bits
    >>> leadingones(np.array([1, 1, 0, 0, 1, 1, 1]))
    5

    # n = 7 and 5 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 0, 1]))
    7

    # n = 7 and 6 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 1]))
    5

    # n = 7 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0, 1, 1]))
    3

    # n = 7 and 6 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 1]))
    5

    # n = 7 and 7 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 8 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 0, 1, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 1, 0, 0, 0]))
    8

    # n = 8 and 1 true bit
    >>> leadingones(np.array([1, 0, 0, 0, 0, 0, 0, 0]))
    7

    # n = 8 and 2 true bits
    >>> leadingones(np.array([0, 0, 0, 1, 0, 0, 0, 1]))
    8

    # n = 8 and 2 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 0, 0, 0, 1]))
    8

    # n = 8 and 2 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 0, 0, 0, 0]))
    8

    # n = 8 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 1, 0, 1, 0]))
    7

    # n = 8 and 3 true bits
    >>> leadingones(np.array([1, 0, 1, 0, 1, 0, 0, 0]))
    7

    # n = 8 and 3 true bits
    >>> leadingones(np.array([0, 0, 0, 1, 0, 0, 1, 1]))
    8

    # n = 8 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 0, 0, 0, 1, 1]))
    6

    # n = 8 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 0, 1, 0, 0]))
    6

    # n = 8 and 4 true bits
    >>> leadingones(np.array([0, 0, 1, 0, 1, 1, 1, 0]))
    8

    # n = 8 and 5 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0, 0, 1, 0]))
    4

    # n = 8 and 5 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 0, 0, 1, 1]))
    7

    # n = 8 and 5 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 1, 1, 0, 1]))
    8

    # n = 8 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0, 1, 0, 1]))
    4

    # n = 8 and 6 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 0, 1]))
    6

    # n = 8 and 6 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 0, 1]))
    6

    # n = 8 and 7 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 1, 1, 1]))
    8

    # n = 8 and 7 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 1, 1]))
    6

    # n = 8 and 7 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 0, 1, 1]))
    3

    # n = 8 and 8 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 9 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0]))
    9

    # n = 9 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 0, 1, 0, 0, 0]))
    9

    # n = 9 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0]))
    9

    # n = 9 and 1 true bit
    >>> leadingones(np.array([0, 0, 1, 0, 0, 0, 0, 0, 0]))
    9

    # n = 9 and 2 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 1, 0, 0, 1]))
    9

    # n = 9 and 2 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 1, 1, 0, 0, 0]))
    9

    # n = 9 and 2 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 0, 0, 1, 0, 0]))
    8

    # n = 9 and 3 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 1, 0, 0, 0, 1]))
    9

    # n = 9 and 3 true bits
    >>> leadingones(np.array([0, 1, 1, 0, 0, 1, 0, 0, 0]))
    9

    # n = 9 and 3 true bits
    >>> leadingones(np.array([0, 1, 1, 0, 0, 0, 0, 0, 1]))
    9

    # n = 9 and 4 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 1, 0, 0, 1, 1]))
    9

    # n = 9 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 0, 1, 0, 0, 1, 0]))
    7

    # n = 9 and 4 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 0, 0, 1, 0, 0]))
    9

    # n = 9 and 5 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 1, 1, 0, 1, 0]))
    9

    # n = 9 and 5 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 1, 0, 0, 1, 1]))
    9

    # n = 9 and 5 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 1, 0, 0, 1, 1]))
    8

    # n = 9 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 1, 0, 0, 1]))
    6

    # n = 9 and 6 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 0, 0]))
    8

    # n = 9 and 6 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 1, 1, 0, 0, 1]))
    7

    # n = 9 and 7 true bits
    >>> leadingones(np.array([1, 1, 0, 1, 0, 1, 1, 1, 1]))
    7

    # n = 9 and 7 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 0, 1, 1]))
    8

    # n = 9 and 7 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 0, 1, 1, 1]))
    8

    # n = 9 and 8 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 1, 1, 1, 1]))
    6

    # n = 9 and 8 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1]))
    8

    # n = 9 and 8 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1]))
    8

    # n = 9 and 9 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0

    # n = 10 and 0 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 1, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 1 true bit
    >>> leadingones(np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0]))
    9

    # n = 10 and 1 true bit
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1]))
    10

    # n = 10 and 2 true bits
    >>> leadingones(np.array([0, 1, 1, 0, 0, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 2 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 0, 0, 0, 0, 0, 0]))
    10

    # n = 10 and 2 true bits
    >>> leadingones(np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1]))
    10

    # n = 10 and 3 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 0, 0, 0, 1, 0, 0]))
    10

    # n = 10 and 3 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 1, 0, 0, 1, 0, 0]))
    9

    # n = 10 and 3 true bits
    >>> leadingones(np.array([0, 1, 0, 0, 0, 1, 0, 0, 0, 1]))
    10

    # n = 10 and 4 true bits
    >>> leadingones(np.array([1, 1, 0, 0, 0, 1, 0, 0, 1, 0]))
    8

    # n = 10 and 4 true bits
    >>> leadingones(np.array([1, 0, 0, 1, 0, 1, 1, 0, 0, 0]))
    9

    # n = 10 and 4 true bits
    >>> leadingones(np.array([1, 0, 0, 0, 1, 1, 0, 0, 0, 1]))
    9

    # n = 10 and 5 true bits
    >>> leadingones(np.array([0, 0, 0, 1, 1, 1, 0, 0, 1, 1]))
    10

    # n = 10 and 5 true bits
    >>> leadingones(np.array([0, 0, 1, 1, 0, 1, 1, 1, 0, 0]))
    10

    # n = 10 and 5 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 0, 0, 1, 0, 0]))
    10

    # n = 10 and 6 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 0, 0, 0]))
    9

    # n = 10 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 0, 1, 1, 0, 0]))
    7

    # n = 10 and 6 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 0, 1, 0, 0, 1]))
    7

    # n = 10 and 7 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 1, 0, 1, 1, 0]))
    10

    # n = 10 and 7 true bits
    >>> leadingones(np.array([0, 1, 0, 1, 1, 1, 1, 1, 1, 0]))
    10

    # n = 10 and 7 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 0, 0, 1, 0, 1]))
    5

    # n = 10 and 8 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 1, 0, 1]))
    9

    # n = 10 and 8 true bits
    >>> leadingones(np.array([1, 1, 1, 0, 1, 1, 0, 1, 1, 1]))
    7

    # n = 10 and 8 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0, 0, 1, 1, 1, 1]))
    6

    # n = 10 and 9 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 0, 1, 1, 1, 1, 1]))
    6

    # n = 10 and 9 true bits
    >>> leadingones(np.array([1, 0, 1, 1, 1, 1, 1, 1, 1, 1]))
    9

    # n = 10 and 9 true bits
    >>> leadingones(np.array([0, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    10

    # n = 10 and 10 true bits
    >>> leadingones(np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))
    0
    """
    xlen: Final[int] = len(x)
    for i in range(xlen):
        if not x[i]:
            return xlen - i
    return 0


class LeadingOnes(BitStringProblem):
    """Maximize the number of leading ones in a bit string."""

    def __init__(self, n: int) -> None:
        """
        Initialize the leading ones objective function.

        :param n: the dimension of the problem

        >>> LeadingOnes(55).n
        55
        >>> LeadingOnes(4).evaluate(np.array([True, True, True, False]))
        1
        """
        super().__init__(n)
        self.evaluate = leadingones  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the leadingones objective function.

        :return: `leadingones_` + lenth of string

        >>> LeadingOnes(10)
        leadingones_10
        """
        return f"leadingones_{self.n}"

    @classmethod
    def default_instances(
            cls: type, scale_min: int = 2, scale_max: int = 4096) \
            -> Iterator[Callable[[], "LeadingOnes"]]:
        """
        Get the 163 default instances of the :class:`LeadingOnes` problem.

        :param scale_min: the minimum permitted scale, by default `2`
        :param scale_max: the maximum permitted scale, by default `4096`
        :returns: a sequence of default :class:`LeadingOnes` instances

        >>> len(list(LeadingOnes.default_instances()))
        163

        >>> [x() for x in LeadingOnes.default_instances()]
        [leadingones_2, leadingones_3, leadingones_4, leadingones_5, \
leadingones_6, leadingones_7, leadingones_8, leadingones_9, leadingones_10, \
leadingones_11, leadingones_12, leadingones_13, leadingones_14, \
leadingones_15, leadingones_16, leadingones_17, leadingones_18, \
leadingones_19, leadingones_20, leadingones_21, leadingones_22, \
leadingones_23, leadingones_24, leadingones_25, leadingones_26, \
leadingones_27, leadingones_28, leadingones_29, leadingones_30, \
leadingones_31, leadingones_32, leadingones_33, leadingones_36, \
leadingones_40, leadingones_41, leadingones_42, leadingones_44, \
leadingones_48, leadingones_49, leadingones_50, leadingones_55, \
leadingones_59, leadingones_60, leadingones_64, leadingones_66, \
leadingones_70, leadingones_77, leadingones_79, leadingones_80, \
leadingones_81, leadingones_85, leadingones_88, leadingones_90, \
leadingones_96, leadingones_99, leadingones_100, leadingones_107, \
leadingones_111, leadingones_121, leadingones_125, leadingones_128, \
leadingones_144, leadingones_149, leadingones_169, leadingones_170, \
leadingones_192, leadingones_196, leadingones_199, leadingones_200, \
leadingones_222, leadingones_225, leadingones_243, leadingones_256, \
leadingones_269, leadingones_289, leadingones_300, leadingones_324, \
leadingones_333, leadingones_341, leadingones_343, leadingones_359, \
leadingones_361, leadingones_384, leadingones_400, leadingones_441, \
leadingones_444, leadingones_479, leadingones_484, leadingones_500, \
leadingones_512, leadingones_529, leadingones_555, leadingones_576, \
leadingones_600, leadingones_625, leadingones_641, leadingones_666, \
leadingones_676, leadingones_682, leadingones_700, leadingones_729, \
leadingones_768, leadingones_777, leadingones_784, leadingones_800, \
leadingones_841, leadingones_857, leadingones_888, leadingones_900, \
leadingones_961, leadingones_999, leadingones_1000, leadingones_1024, \
leadingones_1089, leadingones_1111, leadingones_1151, leadingones_1156, \
leadingones_1225, leadingones_1296, leadingones_1365, leadingones_1369, \
leadingones_1444, leadingones_1521, leadingones_1536, leadingones_1543, \
leadingones_1600, leadingones_1681, leadingones_1764, leadingones_1849, \
leadingones_1936, leadingones_2000, leadingones_2025, leadingones_2048, \
leadingones_2063, leadingones_2116, leadingones_2187, leadingones_2209, \
leadingones_2222, leadingones_2304, leadingones_2401, leadingones_2500, \
leadingones_2601, leadingones_2704, leadingones_2730, leadingones_2753, \
leadingones_2809, leadingones_2916, leadingones_3000, leadingones_3025, \
leadingones_3072, leadingones_3125, leadingones_3136, leadingones_3249, \
leadingones_3333, leadingones_3364, leadingones_3481, leadingones_3600, \
leadingones_3671, leadingones_3721, leadingones_3844, leadingones_3969, \
leadingones_4000, leadingones_4096]
        """
        return cast("Iterator[Callable[[], LeadingOnes]]",
                    super().default_instances(  # type: ignore
                        scale_min, scale_max))



"""
A fitness combining the rank of a solution with the iteration of its creation.

This fitness assignment process is compatible with the simple (mu+lambda)
Evolutionary Algorithm implemented in :class:`~moptipy.algorithms.so.ea.EA`.
It will assign a better fitness to a solution which has a better objective
value. Ties will be broken based on the iteration counter
:attr:`~moptipy.algorithms.so.record.Record.it` of the solution records
:class:`~moptipy.algorithms.so.record.Record`.
"""

from math import inf
from typing import Final

from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


# start book
class RankAndIteration(Fitness):
    """
    A fitness joining objective rank and creation iteration.

    The fitness assignment strategy will use two pieces of information to
    determine the fitness of a solution:

    1. the rank the solutions by their objective values
       (:attr:`~moptipy.algorithms.so.record.Record.f`). If two solutions have
       the same fitness, they get the same rank, but the next-worst solution
       will then get a rank with is larger by 2.
    2. the iteration index (:attr:`~moptipy.algorithms.so.record.Record.it`)
       relative to the maximum and minimum iteration index.

    It will multiply the rank of the solution with the range of the iteration
    index in the population and then add the maximum iteration index minus the
    iteration index of the solution, i.e.,

    `fitness(x) = rank(f(x)) * (max_it - min_it + 1)  +  (max_it - it(x))`

    This way, better solutions receive better fitness and ties are broken such
    that younger solutions (with higher iteration index) are preferred.
    In combination with best selection
    (:class:`moptipy.algorithms.modules.selections.best.Best`),
    this replicates the behavior of our simple (mu+lambda) Evolutionary
    Algorithm (:class:`~moptipy.algorithms.so.ea.EA`).
    """

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the rank and iteration fitness.

        :param p: the list of records
        :param random: ignored
        """
        min_it: int = 9_223_372_036_854_775_808  # minimum iteration index
        max_it: int = -1  # the maximum iteration index

        # In the first iteration, we assign objective value as fitness
        # (for sorting) and get the bounds of the iteration indices.
        for rec in p:  # iterate over list p
            rec.fitness = rec.f  # set f as fitness for sorting
            it: int = rec.it  # get iteration index from record
            min_it = min(min_it, it)
            max_it = max(max_it, it)
        p.sort()  # sort based on objective values

        it_range: Final[int] = max_it - min_it + 1  # range of it index
        rank: int = -1  # the variable for storing the current rank
        last_f: int | float = -inf  # the previous objective value
        for i, rec in enumerate(p):  # iterate over list
            v = rec.fitness  # get the current objective value
            if v > last_f:  # only increase rank if objective f changes
                rank = i + 1  # +1 so smallest-possible fitness is 1
                last_f = v  # remember objective value for comparison
            rec.fitness = (rank * it_range) + max_it - rec.it
# end book

    def __str__(self):
        """
        Get the name of this fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "rankAndIt": always
        """
        return "rankAndIt"



"""
The FFA-based version of the Hill Climber: The FHC.

This algorithm is based on
:class:`~moptipy.algorithms.so.hill_climber.HillClimber`, i.e., the simple
Hill Climber, but uses Frequency Fitness Assignment (FFA) as fitness
assignment process. FFA replaces all objective values with their encounter
frequencies in the selection decisions. The more often an objective value is
encountered, the higher gets its encounter frequency. Therefore, local
optima are slowly receiving worse and worse fitness.

This algorithm is closely related
:class:`~moptipy.algorithms.so.ffa.fea1plus1.FEA1plus1`.
The difference is that it accepts new solutions only if their frequency
fitness is *lower* than the frequency fitness of the current solution.
:class:`~moptipy.algorithms.so.ffa.fea1plus1.FEA1plus1`, on the other hand,
accepts solutions if their frequency fitness is *lower or equal*
than the frequency fitness of the current solution.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*.
   27(4):980-992. August 2023.
   doi: https://doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
7. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, Sarah Louise
   Thomson, and Thomas Weise. Addressing the Traveling Salesperson Problem
   with Frequency Fitness Assignment and Hybrid Algorithms. *Soft Computing.*
   2024. https://dx.doi.org/10.1007/s00500-024-09718-8
"""
from collections import Counter
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


class FHC(Algorithm1):
    """
    The FFA-based version of the Hill Climber.

    This algorithm applies Frequency Fitness Assignment (FFA).
    This means that it does not select solutions based on whether
    they are better or worse. Instead, it selects the solution whose
    objective value has been encountered during the search less often.
    The word "best" therefore is not used in the traditional sense, i.e.,
    that one solution is better than another one terms of its objective
    value. Instead, the current best solution is always the one whose
    objective value we have seen the least often.

    This algorithm implementation works best if objective values are
    integers and have lower and upper bounds that are not too far
    away from each other. If there are too many possible different objective
    values, the algorithm degenerates to a random walk.
    """

    def __init__(self, op0: Op0, op1: Op1, log_h_tbl: bool = False) -> None:
        """
        Create the FHC.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param log_h_tbl: should we log the H table?
        """
        super().__init__("fhc", op0, op1)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        #: True if we should log the H table, False otherwise
        self.__log_h_tbl: Final[bool] = log_h_tbl

    def solve(self, process: Process) -> None:
        """
        Apply the FHC to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        cur_x = process.create()  # record for current solution
        new_x = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, cur_x)  # Create 1 solution randomly and
        cur_f: int | float = evaluate(cur_x) + ofs  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, cur_x)  # new_x = neighbor of cur_x
            new_f: int | float = evaluate(new_x) + ofs

            h[new_f] += 1  # type: ignore  # Increase the frequency
            h[cur_f] += 1  # type: ignore  # of new_f and cur_f.
            if h[new_f] < h[cur_f]:  # type: ignore
                cur_f = new_f  # Store its objective value.
                cur_x, new_x = new_x, cur_x  # Swap best and new.

        if not self.__log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[cur_f] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[cur_f] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table
